__author__ = "Simon Nilsson"


import glob
import os
from copy import deepcopy
from typing import Dict, List, Optional, Union

import cv2
import numpy as np
from PIL import Image

from simba.mixins.config_reader import ConfigReader
from simba.mixins.plotting_mixin import PlottingMixin
from simba.mixins.train_model_mixin import TrainModelMixin
from simba.utils.checks import (check_file_exist_and_readable, check_float,
                                check_int)
from simba.utils.data import create_color_palette
from simba.utils.enums import ConfigKey, Dtypes, Formats, TagNames
from simba.utils.errors import NoSpecifiedOutputError
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (get_fn_ext, get_video_meta_data,
                                    read_config_entry, read_df)


class PlotSklearnResultsSingleCore(ConfigReader, TrainModelMixin, PlottingMixin):
    """
    Plot classification results overlays on videos. Results are stored in the
    `project_folder/frames/output/sklearn_results` directory of the SimBA project.

    .. note::
       For improved run-time, see :meth:`simba.plotting.plot_clf_results_mp.PlotSklearnResultsMultiProcess` for multiprocess class.
       Scikit visualization documentation <https://github.com/sgoldenlab/simba/blob/master/docs/tutorial.md#step-10-sklearn-visualization__.

    .. image:: _static/img/sklearn_visualization.gif
       :width: 600
       :align: center


    :param str config_path: path to SimBA project config file in Configparser format
    :param bool rotate: If True, the output video will be rotated 90 degrees from the input.
    :param bool video_setting: If True, SimBA will create compressed videos.
    :param bool frame_setting: If True, SimBA will create individual frames.
    :param Optional[dict] text_settings:
    :param str video_file_path: path to video file to create classification visualizations for.

    Examples
    ----------
    >>> clf_plotter = PlotSklearnResultsSingleCore(config_path='MyProjectConfig', video_setting=True, frame_setting=False, rotate=False, video_file_path='VideoPath')
    >>> clf_plotter.run()
    """

    def __init__(
        self,
        config_path: str,
        video_setting: bool,
        frame_setting: bool,
        text_settings: Optional[Union[Dict[str, float], bool]] = False,
        video_file_path: Optional[List] = None,
        rotate: bool = False,
        print_timers: bool = True,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        TrainModelMixin.__init__(self)
        PlottingMixin.__init__(self)
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )

        if (not video_setting) and (not frame_setting):
            raise NoSpecifiedOutputError(
                msg="Please choose to create a video and/or frames. SimBA found that you ticked neither video and/or frames",
                source=self.__class__.__name__,
            )
        self.video_file_path, self.print_timers, self.text_settings = (
            video_file_path,
            print_timers,
            text_settings,
        )
        self.video_setting, self.frame_setting = video_setting, frame_setting
        if video_file_path is not None:
            check_file_exist_and_readable(os.path.join(self.video_dir, video_file_path))
        if not os.path.exists(self.sklearn_plot_dir):
            os.makedirs(self.sklearn_plot_dir)
        self.pose_threshold = read_config_entry(
            self.config,
            ConfigKey.THRESHOLD_SETTINGS.value,
            ConfigKey.SKLEARN_BP_PROB_THRESH.value,
            Dtypes.FLOAT.value,
            0.00,
        )
        self.clr_lst = create_color_palette(
            pallete_name="Set1", increments=self.clf_cnt
        )
        self.files_found = glob.glob(self.machine_results_dir + "/*." + self.file_type)
        self.model_dict = self.get_model_info(self.config, self.clf_cnt)
        self.fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        self.rotate = rotate
        self.a = np.deg2rad(90)
        print(f"Processing {str(len(self.files_found))} videos...")

    def __get_print_settings(self):
        if self.text_settings is False:
            self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
                60,
                12,
                1500,
                1.1,
            )
            self.max_dim = max(
                self.video_meta_data["width"], self.video_meta_data["height"]
            )
            self.circle_scale = int(self.radius_scale / (self.res_scale / self.max_dim))
            self.font_size = float(self.font_scale / (self.res_scale / self.max_dim))
            self.spacing_scale = int(self.space_scale / (self.res_scale / self.max_dim))
            self.text_thickness = 2
        else:
            check_float(name="ERROR: TEXT SIZE", value=self.text_settings["font_size"])
            check_int(name="ERROR: SPACE SIZE", value=self.text_settings["space_size"])
            check_int(
                name="ERROR: TEXT THICKNESS", value=self.text_settings["text_thickness"]
            )
            check_int(
                name="ERROR: CIRCLE SIZE", value=self.text_settings["circle_size"]
            )
            self.font_size = float(self.text_settings["font_size"])
            self.spacing_scale = int(self.text_settings["space_size"])
            self.text_thickness = int(self.text_settings["text_thickness"])
            self.circle_scale = int(self.text_settings["circle_size"])

    def create_visualizations(self):
        _, self.video_name, _ = get_fn_ext(self.file_path)
        self.data_df = read_df(self.file_path, self.file_type).reset_index(drop=True)
        self.video_settings, _, self.fps = self.read_video_info(
            video_name=self.video_name
        )
        self.video_path = self.find_video_of_file(self.video_dir, self.video_name)
        self.cap = cv2.VideoCapture(self.video_path)
        self.save_path = os.path.join(self.sklearn_plot_dir, self.video_name + ".mp4")
        self.video_meta_data = get_video_meta_data(self.video_path)
        height, width = deepcopy(self.video_meta_data["height"]), deepcopy(
            self.video_meta_data["width"]
        )
        if self.frame_setting:
            self.video_frame_dir = os.path.join(self.sklearn_plot_dir, self.video_name)
            if not os.path.exists(self.video_frame_dir):
                os.makedirs(self.video_frame_dir)
        if self.rotate:
            self.video_meta_data["height"], self.video_meta_data["width"] = (
                width,
                height,
            )
        self.writer = cv2.VideoWriter(
            self.save_path,
            self.fourcc,
            self.fps,
            (self.video_meta_data["width"], self.video_meta_data["height"]),
        )
        self.__get_print_settings()
        self.video_model_dict = deepcopy(self.model_dict)
        for model in self.video_model_dict:
            self.video_model_dict[model]["time"] = 0

        row_n = 0
        while self.cap.isOpened():
            ret, self.frame = self.cap.read()
            try:
                if ret:
                    self.id_flag_cords = {}
                    for animal_name, animal_data in self.animal_bp_dict.items():
                        animal_clr = animal_data["colors"]
                        ID_flag = False
                        for bp_no in range(len(animal_data["X_bps"])):
                            bp_clr = animal_clr[bp_no]
                            x_bp, y_bp = (
                                animal_data["X_bps"][bp_no],
                                animal_data["Y_bps"][bp_no],
                            )
                            p_bp = x_bp[:-2] + "_p"
                            bp_cords = self.data_df.loc[row_n, [x_bp, y_bp, p_bp]]
                            if bp_cords[p_bp] > self.pose_threshold:
                                cv2.circle(
                                    self.frame,
                                    (int(bp_cords[x_bp]), int(bp_cords[y_bp])),
                                    0,
                                    bp_clr,
                                    self.circle_scale,
                                )
                                if ("centroid" in x_bp.lower()) or (
                                    "center" in x_bp.lower()
                                ):
                                    self.id_flag_cords[animal_name] = (
                                        int(bp_cords[x_bp]),
                                        int(bp_cords[y_bp]),
                                    )
                                    ID_flag = True

                        if not ID_flag:
                            self.id_flag_cords[animal_name] = (
                                int(bp_cords[x_bp]),
                                int(bp_cords[y_bp]),
                            )

                    for animal_name, animal_cords in self.id_flag_cords.items():
                        cv2.putText(
                            self.frame,
                            animal_name,
                            animal_cords,
                            self.font,
                            self.font_size,
                            self.animal_bp_dict[animal_name]["colors"][0],
                            2,
                        )

                    if self.rotate:
                        self.frame = np.array(
                            Image.fromarray(self.frame).rotate(
                                90, Image.BICUBIC, expand=True
                            )
                        )
                    if self.print_timers:
                        cv2.putText(
                            self.frame,
                            str("Timers"),
                            (
                                10,
                                (
                                    (
                                        self.video_meta_data["height"]
                                        - self.video_meta_data["height"]
                                    )
                                    + self.spacing_scale
                                ),
                            ),
                            self.font,
                            self.font_size,
                            (0, 255, 0),
                            self.text_thickness,
                        )
                    self.add_spacer = 2
                    for model_no, model_info in self.video_model_dict.items():
                        frame_results = self.data_df.loc[
                            row_n, model_info["model_name"]
                        ]
                        self.video_model_dict[model_no]["frame_results"] = frame_results
                        self.video_model_dict[model_no]["time"] += (
                            frame_results / self.fps
                        )
                        if self.print_timers:
                            cv2.putText(
                                self.frame,
                                model_info["model_name"]
                                + " "
                                + str(round(self.video_model_dict[model_no]["time"], 2))
                                + str("s"),
                                (
                                    10,
                                    (
                                        self.video_meta_data["height"]
                                        - self.video_meta_data["height"]
                                    )
                                    + self.spacing_scale * self.add_spacer,
                                ),
                                self.font,
                                self.font_size,
                                (255, 0, 0),
                                self.text_thickness,
                            )
                            self.add_spacer += 1
                    cv2.putText(
                        self.frame,
                        str("Ensemble prediction"),
                        (
                            10,
                            (
                                self.video_meta_data["height"]
                                - self.video_meta_data["height"]
                            )
                            + self.spacing_scale * self.add_spacer,
                        ),
                        self.font,
                        self.font_size,
                        (0, 255, 0),
                        self.text_thickness,
                    )
                    self.add_spacer += 1

                    for model_cnt, model_info in self.video_model_dict.items():
                        if self.video_model_dict[model_cnt]["frame_results"] == 1:
                            cv2.putText(
                                self.frame,
                                model_info["model_name"],
                                (
                                    10,
                                    (
                                        self.video_meta_data["height"]
                                        - self.video_meta_data["height"]
                                    )
                                    + self.spacing_scale * self.add_spacer,
                                ),
                                self.font,
                                self.font_size,
                                self.clr_lst[model_cnt],
                                self.text_thickness,
                            )
                            self.add_spacer += 1
                    if self.video_setting:
                        self.writer.write(self.frame)
                    if self.frame_setting:
                        frame_save_name = os.path.join(
                            self.video_frame_dir, str(row_n) + ".png"
                        )
                        cv2.imwrite(frame_save_name, self.frame)
                    print(
                        f'Frame: {row_n} / {self.video_meta_data["frame_count"]}. Video: {self.video_name} ({self.file_cnt + 1}/{len(self.files_found)})'
                    )
                    row_n += 1

                else:
                    print("Video {} saved...".format(self.video_name))
                    self.cap.release()
                    self.writer.release()
            except KeyError as e:
                print(e.args, e)
                print(
                    "SIMBA INDEX WARNING: Some frames appears to be missing in the dataframe and could not be created"
                )
                print("Video {} saved...".format(self.video_name))
                self.cap.release()
                self.writer.release()

    def run(self):
        if self.video_file_path is None:
            for file_cnt, file_path in enumerate(self.files_found):
                self.file_cnt, self.file_path = file_cnt, file_path
                self.create_visualizations()
        else:
            self.file_cnt, file_path = 0, self.video_file_path
            _, file_name, _ = get_fn_ext(file_path)
            self.file_path = os.path.join(
                self.machine_results_dir, file_name + "." + self.file_type
            )
            self.files_found = [self.file_path]
            check_file_exist_and_readable(self.file_path)
            self.create_visualizations()

        self.timer.stop_timer()
        stdout_success(
            msg="All visualizations created in project_folder/frames/output/sklearn_results directory",
            elapsed_time=self.timer.elapsed_time_str,
            source=self.__class__.__name__,
        )


# test = PlotSklearnResultsSingleCore(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini',
#                                       video_setting=True,
#                                       frame_setting=False,
#                                       video_file_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1.avi',
#                                       print_timers=True,
#                                       text_settings=False,
#                                       rotate=False)
# test.run()



__author__ = "Simon Nilsson"

import os
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.directing_animals_visualizer import \
    DirectingOtherAnimalsVisualizer
from simba.plotting.directing_animals_visualizer_mp import \
    DirectingOtherAnimalsVisualizerMultiprocess
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, DropDownMenu
from simba.utils.enums import Formats, Keys, Links, Paths
from simba.utils.errors import AnimalNumberError
from simba.utils.read_write import get_file_name_info_in_directory


class DirectingOtherAnimalsVisualizerPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        if self.animal_cnt == 1:
            raise AnimalNumberError(
                msg="Cannot visualize directionality between animals in a 1 animal project.",
                source=self.__class__.__name__,
            )
        PopUpMixin.__init__(self, title="CREATE ANIMAL DIRECTION VIDEOS")
        self.color_lst = list(self.colors_dict.keys())
        self.color_lst.insert(0, "Random")
        self.size_lst = list(range(1, 11))
        self.data_path = os.path.join(self.project_path, Paths.OUTLIER_CORRECTED.value)
        self.files_found_dict = get_file_name_info_in_directory(
            directory=self.data_path, file_type=self.file_type
        )

        self.show_pose_var = BooleanVar(value=True)
        self.highlight_direction_endpoints_var = BooleanVar(value=True)
        self.merge_directionality_lines_var = BooleanVar(value=False)
        self.multiprocess_var = BooleanVar(value=False)

        self.style_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="STYLE SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.DIRECTING_ANIMALS_PLOTS.value,
        )
        self.show_pose_cb = Checkbutton(
            self.style_settings_frm,
            text="Show pose-estimated body-parts",
            variable=self.show_pose_var,
        )
        self.highlight_direction_endpoints_cb = Checkbutton(
            self.style_settings_frm,
            text="Highlight direction end-points",
            variable=self.highlight_direction_endpoints_var,
        )
        self.merge_directionality_lines_cb = Checkbutton(
            self.style_settings_frm,
            text="Polyfill direction lines",
            variable=self.merge_directionality_lines_var,
        )
        self.direction_clr_dropdown = DropDownMenu(
            self.style_settings_frm, "Direction color:", self.color_lst, "16"
        )
        self.pose_size_dropdown = DropDownMenu(
            self.style_settings_frm, "Pose circle size:", self.size_lst, "16"
        )
        self.line_thickness = DropDownMenu(
            self.style_settings_frm, "Line thickness:", self.size_lst, "16"
        )
        self.line_thickness.setChoices(choice=4)
        self.pose_size_dropdown.setChoices(choice=3)
        self.direction_clr_dropdown.setChoices(choice="Random")
        multiprocess_cb = Checkbutton(
            self.style_settings_frm,
            text="Multi-process (faster)",
            variable=self.multiprocess_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.multiprocess_var,
                dropdown_menus=[self.multiprocess_dropdown],
            ),
        )
        self.multiprocess_dropdown = DropDownMenu(
            self.style_settings_frm, "CPU cores:", list(range(2, self.cpu_cnt)), "12"
        )
        self.multiprocess_dropdown.setChoices(2)
        self.multiprocess_dropdown.disable()

        self.run_frm = LabelFrame(
            self.main_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_frm = LabelFrame(
            self.run_frm,
            text="SINGLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_btn = Button(
            self.run_single_video_frm,
            text="Create single video",
            fg="blue",
            command=lambda: self.__create_directionality_plots(multiple_videos=False),
        )
        self.single_video_dropdown = DropDownMenu(
            self.run_single_video_frm,
            "Video:",
            list(self.files_found_dict.keys()),
            "12",
        )
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])
        self.run_multiple_videos = LabelFrame(
            self.run_frm,
            text="MULTIPLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_multiple_video_btn = Button(
            self.run_multiple_videos,
            text="Create multiple videos ({} video(s) found)".format(
                str(len(list(self.files_found_dict.keys())))
            ),
            fg="blue",
            command=lambda: self.__create_directionality_plots(multiple_videos=True),
        )

        self.style_settings_frm.grid(row=0, column=0, sticky=NW)
        self.show_pose_cb.grid(row=0, column=0, sticky=NW)
        self.highlight_direction_endpoints_cb.grid(row=1, column=0, sticky=NW)
        self.merge_directionality_lines_cb.grid(row=2, column=0, sticky=NW)
        self.direction_clr_dropdown.grid(row=3, column=0, sticky=NW)
        self.pose_size_dropdown.grid(row=4, column=0, sticky=NW)
        self.line_thickness.grid(row=5, column=0, sticky=NW)
        multiprocess_cb.grid(row=6, column=0, sticky=NW)
        self.multiprocess_dropdown.grid(row=6, column=1, sticky=NW)

        self.run_frm.grid(row=1, column=0, sticky=NW)
        self.run_single_video_frm.grid(row=0, column=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)
        self.run_multiple_videos.grid(row=1, column=0, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, column=0, sticky=NW)

    def __create_directionality_plots(self, multiple_videos: bool):
        if multiple_videos:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [
                self.files_found_dict[self.single_video_dropdown.getChoices()]
            ]

        style_attr = {
            "Show_pose": self.show_pose_var.get(),
            "Pose_circle_size": int(self.pose_size_dropdown.getChoices()),
            "Direction_color": self.direction_clr_dropdown.getChoices(),
            "Direction_thickness": int(self.line_thickness.getChoices()),
            "Highlight_endpoints": self.highlight_direction_endpoints_var.get(),
            "Polyfill": self.merge_directionality_lines_var.get(),
        }

        for data_path in data_paths:
            if not self.multiprocess_var.get():
                directing_other_animal_visualizer = DirectingOtherAnimalsVisualizer(
                    config_path=self.config_path,
                    data_path=data_path,
                    style_attr=style_attr,
                )
            else:
                directing_other_animal_visualizer = (
                    DirectingOtherAnimalsVisualizerMultiprocess(
                        config_path=self.config_path,
                        data_path=data_path,
                        style_attr=style_attr,
                        core_cnt=int(self.multiprocess_dropdown.getChoices()),
                    )
                )
            directing_other_animal_visualizer.run()


# _ = DirectingOtherAnimalsVisualizerPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini')



import os
import random
from typing import Union

import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.utils.printing import stdout_success


class GibbsSampler(ConfigReader):
    """
    Gibbs sampling for finding "motifs" in categorical sequences.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter pd.DataFrame data: Dataframe where each row represents videos and each column represents behavior sequence.
    :parameter float pseudo_number: Small error value for fuzzy search. Default: 0.001.
    :parameter int sequence_length: The length of the motif sequence searched for.
    :parameter int iterations: Number of iterations per epoch. Default: 1000.
    :parameter int epochs: Number of epochs of iterations. Default: 5.
    :parameter int plateau_val: Terminate epoch when error rate has remained unchanged for ``plateau_val`` count of interations.

    :example:
    >>> df = pd.read_csv('/tests/data/gibbs_sample_data/gibbs_sample_ver_7.csv', index_col=0)
    >>> gibbs_sampler = GibbsSampler(data=df)
    >>> gibbs_sampler.run()


    :References:
    .. [1] Lawrence et.al, Detecting Subtle Sequence Signals: a Gibbs Sampling Strategy for Multiple Alignment,
           Science, vol. 262, pp. 208-214, 1993.
    .. [2] `Great YouTube course by Xiaole Shirley Liu <https://www.youtube.com/watch?v=NRjhfyXWHuQ>`_.
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        data: pd.DataFrame,
        pseudo_number: int = 0.001,
        sequence_length: int = 4,
        iterations: int = 1500,
        epochs: int = 2,
        stop_val: float = 0.001,
        plateau_val: int = 50,
    ):
        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        self.unique_vals = sorted(list(pd.unique(data.values.ravel("K"))))
        self.target_probability = 1 * sequence_length + (
            pseudo_number * (sequence_length + 1)
        )

        self.holdout_cols = ["H_{}".format(i) for i in range(0, sequence_length)]
        self.non_holdout_cols = ["C_{}".format(i) for i in range(sequence_length)]
        self.output_cols = ["Behavior_{}".format(i + 1) for i in range(sequence_length)]

        self.data, self.pseudo_num, self.sequence_len = (
            data,
            pseudo_number,
            sequence_length,
        )
        self.iterations, self.stop_val = [iterations] * epochs, stop_val
        self.plateau_val = plateau_val
        self.summary_df = pd.DataFrame(columns=self.output_cols)
        self.data.columns = ["F_{}".format(x) for x in range(len(self.data.columns))]
        self.save_path = os.path.join(
            self.logs_path, f"gibbs_sampling_results_{self.datetime}.csv"
        )

    def __get_start_positions(self):
        start_positions = []
        for val in range(len(self.data)):
            start_positions.append(
                random.randint(0, len(self.data.columns) - self.sequence_len)
            )
        return start_positions

    def __make_probability_table(self, df: pd.DataFrame, unique_vals: np.ndarray):
        prob_df = pd.DataFrame(columns=unique_vals)
        for column in df:
            occurance_lst = []
            for value in unique_vals:
                occurance_lst.append(
                    (len(df[df[column] == value]) + self.pseudo_num) / len(df[column])
                    + self.pseudo_num
                )
            prob_df.loc[len(prob_df)] = occurance_lst
        return prob_df

    def __iterate_through_possible_seq_in_hold_out(
        self,
        hoObs: pd.DataFrame,
        bgDf: pd.DataFrame,
        probDf: pd.DataFrame,
        start_hold_no: int,
    ):
        df_holdout_seqs = pd.DataFrame(columns=self.holdout_cols + ["Prob_weight"])
        for startv in range(0, len(hoObs.columns) - self.sequence_len + 1):
            cols = [
                "F_{}".format(ind) for ind in range(startv, startv + self.sequence_len)
            ]
            seq = hoObs.loc[:, cols]
            ptot = 1
            for ind in range(self.sequence_len):
                prob_table_cols = seq.loc[start_hold_no][ind]
                ptot *= probDf.loc[ind][prob_table_cols] / bgDf.loc[prob_table_cols]
            seq["Prob_weight"] = ptot
            seq.columns = self.holdout_cols + ["Prob_weight"]
            df_holdout_seqs = df_holdout_seqs.append(seq, ignore_index=True)
        return df_holdout_seqs

    def __sum_results(self, full_sequence_set, summary_df):
        for i in range(self.sequence_len):
            full_sequence_set = full_sequence_set.rename(
                columns={f"C_{i}": self.output_cols[i]}
            )

        summary_df = pd.concat([summary_df, full_sequence_set], axis=0).reset_index(
            drop=True
        )
        output = (
            summary_df.groupby(summary_df.columns.tolist())
            .size()
            .reset_index()
            .rename(columns={0: "records"})
            .sort_values(by=["records"], ascending=False)
        )
        output["percent"] = output["records"] / output["records"].sum()

        return summary_df, output.reset_index(drop=True)

    def run(self):
        # calculate background constants
        bg_df = self.data.stack().value_counts()
        bg_df = bg_df / bg_df.sum()
        bg_df = bg_df.sort_index()
        full_sequence_set, output = None, None
        prior_error, stable_error_cnt = np.inf, 0

        for iteration_counter, iterations in enumerate(self.iterations):
            holdout_idx = np.random.randint(0, len(self.data) - 2)
            start_positions = self.__get_start_positions()
            for iteration in range(0, iterations):
                hold_out_observation_df = self.data.iloc[[holdout_idx]]
                if iteration == 0:
                    df_it = self.data.drop(hold_out_observation_df.index)
                    idx_lst, seq_lst = [], []
                    for index, row in df_it.iterrows():
                        seq_lst.append(
                            list(
                                row[
                                    start_positions[index] : start_positions[index]
                                    + self.sequence_len
                                ]
                            )
                        )
                        idx_lst.append(index)
                    non_holdout_df = pd.DataFrame(
                        seq_lst, columns=self.non_holdout_cols, index=idx_lst
                    )
                else:
                    non_holdout_df = full_sequence_set.loc[
                        full_sequence_set.index != holdout_idx
                    ]

                probability_df = self.__make_probability_table(
                    non_holdout_df, self.unique_vals
                )
                df_holdout_seqs = self.__iterate_through_possible_seq_in_hold_out(
                    hold_out_observation_df, bg_df, probability_df, holdout_idx
                )
                new_seq = df_holdout_seqs.sample(1, weights="Prob_weight")
                new_seq = new_seq[self.holdout_cols]
                new_seq.columns = self.non_holdout_cols
                new_seq.index = [holdout_idx]
                full_sequence_set = pd.concat([new_seq, non_holdout_df])
                error = self.target_probability - probability_df.max(axis=1).sum()
                # self.summary_df, output = self.__sum_results(full_sequence_set, self.summary_df, iteration_counter)
                print(
                    f"Current error: {error}, iteration: {iteration}, epoch: {iteration_counter}"
                )
                if error <= self.stop_val:
                    print(f"Convergence reached in {iteration} iterations")
                    break
                if error == prior_error:
                    stable_error_cnt += 1
                    if stable_error_cnt >= self.plateau_val - 1:
                        print(f"Platea reached {self.plateau_val}")
                        break
                else:
                    stable_error_cnt = 0
                prior_error = error
                if holdout_idx < (len(self.data) - 1):
                    holdout_idx += 1
                else:
                    holdout_idx = 0
            self.summary_df, output = self.__sum_results(
                full_sequence_set, self.summary_df
            )

        output.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Gibbs sampling results saved @{self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )


# df = pd.read_csv('/Users/simon/Desktop/envs/simba_dev/tests/data/gibbs_sample_data/gibbs_sample_7.csv', index_col=0)
# test = GibbsSampler(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', data=df)
# test.run()



__author__ = "Simon Nilsson"

import collections
import os
from tkinter import *

import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, hxtScrollbar
from simba.utils.enums import (ConfigKey, Dtypes, Formats, Keys, Links, Paths,
                               TagNames)
from simba.utils.errors import (InvalidInputError, NoFilesFoundError,
                                ParametersFileError, PermissionError)
from simba.utils.printing import log_event, stdout_success
from simba.utils.read_write import (find_all_videos_in_project, get_fn_ext,
                                    get_video_meta_data, read_config_entry,
                                    read_video_info_csv)
from simba.utils.warnings import DuplicateNamesWarning, InvalidValueWarning
from simba.video_processors.px_to_mm import get_coordinates_nilsson


class VideoInfoTable(ConfigReader, PopUpMixin):
    """
    Create GUI that allows users to modify resolutions, fps, and pixels-per-mm
    interactively of videos within the SimBA project. Data is stored within the project_folder/logs/video_info.csv
    file in the SimBA project.

    :param str config_file_path: path to SimBA project config file in Configparser format

    ..seealso::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario1.md#step-3-set-video-parameters>`__.

    :example:
    >>> video_info_gui = VideoInfoTable(config_path='MyProjectConfig')
    >>> video_info_gui.create_window()
    >>> video_info_gui.main_frm.mainloop()
    """

    def __init__(self, config_path: str):

        ConfigReader.__init__(self, config_path=config_path, read_video_info=False)
        log_event(
            logger_name=str(__class__.__name__),
            log_type=TagNames.CLASS_INIT.value,
            msg=self.create_log_msg_from_init_args(locals=locals()),
        )
        if os.path.isfile(self.video_info_path):
            self.video_info_df = read_video_info_csv(self.video_info_path).reset_index(
                drop=True
            )
        else:
            self.video_info_df = None
        self.distance_mm = read_config_entry(
            self.config,
            ConfigKey.FRAME_SETTINGS.value,
            ConfigKey.DISTANCE_MM.value,
            Dtypes.FLOAT.value,
            0.00,
        )
        self.video_paths = find_all_videos_in_project(videos_dir=self.video_dir)
        self.video_names = []
        for i in self.video_paths:
            _, video_name, _ = get_fn_ext(filepath=i)
            self.video_names.append(video_name)
        self.video_basename_lst = []
        self.max_char_vid_name = len(max(self.video_names, key=len))

    def __check_that_value_is_numeric(
        self, value=None, value_name=None, video_name=None
    ):
        if str(value).isdigit():
            return int(value)
        elif str(value).replace(".", "", 1).isdigit() and str(value).count(".") < 2:
            return float(value)
        else:
            raise ParametersFileError(
                msg=f"The {value_name} setting for video {video_name} is set to {value} in the project_folder/logs/video_info. Please set it to a numeric value.",
                source=self.__class__.__name__,
            )

    def __append_videos_from_video_folder(self):
        for cnt, name in enumerate(self.video_paths):
            _, video_basename, _ = get_fn_ext(name)
            self.video_basename_lst.append(video_basename)
            self.videos[name] = {}
            self.videos[name]["video_idx_lbl"] = Label(
                self.video_frm, text=str(cnt), width=6
            )
            self.videos[name]["video_name_lbl"] = Label(
                self.video_frm, text=video_basename, width=self.max_char_vid_name
            )
            self.videos[name]["video_name_w_ext"] = name
            video_meta = get_video_meta_data(os.path.join(self.video_dir, name))
            self.videos[name]["fps_var"] = IntVar()
            self.videos[name]["fps_var"].set(video_meta["fps"])
            self.videos[name]["fps_entry"] = Entry(
                self.video_frm, width=20, textvariable=self.videos[name]["fps_var"]
            )
            self.videos[name]["width_var"] = IntVar()
            self.videos[name]["width_var"].set(video_meta["width"])
            self.videos[name]["width_entry"] = Entry(
                self.video_frm, width=20, textvariable=self.videos[name]["width_var"]
            )
            self.videos[name]["height_var"] = IntVar()
            self.videos[name]["height_var"].set(video_meta["height"])
            self.videos[name]["height_entry"] = Entry(
                self.video_frm, width=20, textvariable=self.videos[name]["height_var"]
            )
            self.videos[name]["distance"] = StringVar()
            self.videos[name]["distance"].set(self.distance_mm)
            self.videos[name]["distance_entry"] = Entry(
                self.video_frm, width=20, textvariable=self.videos[name]["distance"]
            )
            self.videos[name]["find_dist_btn"] = Button(
                self.video_frm,
                text="Calculate distance",
                fg="black",
                command=lambda k=(
                    self.videos[name]["video_name_w_ext"],
                    self.videos[name]["distance"],
                ): self.__initiate_find_distance(k),
            )
            self.videos[name]["px_mm"] = StringVar()
            self.videos[name]["px_mm"].set(0)
            self.videos[name]["px_mm_entry"] = Entry(
                self.video_frm, width=20, textvariable=self.videos[name]["px_mm"]
            )
            if isinstance(self.video_info_df, pd.DataFrame):
                prior_data = None
                try:
                    prior_data = self.read_video_info(
                        video_name=video_basename, raise_error=False
                    )[0]
                except ParametersFileError:
                    pass
                if prior_data is not None:
                    for value_name, set_name in zip(
                        [
                            "fps",
                            "Resolution_width",
                            "Resolution_height",
                            "Distance_in_mm",
                            "pixels/mm",
                        ],
                        ["fps_var", "width_var", "height_var", "distance", "px_mm"],
                    ):
                        float_val = self.__check_that_value_is_numeric(
                            value=prior_data[value_name].values[0],
                            value_name=value_name,
                            video_name=prior_data["Video"].values[0],
                        )
                        self.videos[name][set_name].set(float_val)

    def __check_for_duplicate_names(self):
        duplicate_video_names = [
            item
            for item, count in collections.Counter(self.video_basename_lst).items()
            if count > 1
        ]
        if len(duplicate_video_names) > 0:
            for video_name in duplicate_video_names:
                print(video_name)
            DuplicateNamesWarning(
                msg=f"SimBA found {str(len(duplicate_video_names))} duplicate video name(s) in your SimBA project. The video(s) with duplicated names are printed above. This can happen if you have imported a video called MyVideo.mp4, and MyVideo.avi. Please avoid non-unique video names with different file extensions / video formats.",
                source=self.__class__.__name__,
            )

    def __check_no_zero_px_per_mm_values(self):
        for index, row in self.video_df.iterrows():
            try:
                px_per_mm_float = float(row["pixels/mm"])
            except:
                raise ParametersFileError(
                    msg=f'The pixels/mm for video {row["Video"]} is not a numeric value. Please set pixels per mm using the "Calculate distance" button before proceding.',
                    source=self.__class__.__name__,
                )
            if px_per_mm_float <= 0:
                InvalidValueWarning(
                    msg=f'The pixels/mm for video {row["Video"]} is set to zero. Please calculate pixels per mm using the "Calculate distance" button before proceding.',
                    source=self.__class__.__name__,
                )

    def create_window(self):
        PopUpMixin.__init__(self, title="VIDEO INFO", size=(1550, 800))
        self.instructions_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="INSTRUCTIONS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.VIDEO_PARAMETERS.value,
        )
        self.intructions_label_1 = Label(
            self.instructions_frm,
            text='1. Enter the known distance (millimeters) in the "DISTANCE IN MM" column. Consider using the "autopopulate" entry box in the main project window you have a lot of videos.',
            font=("Helvetica", 15),
        )
        self.intructions_label_2 = Label(
            self.instructions_frm,
            text='2. Click on "Calculate distance" button(s) to calculate pixels/mm for each video.',
            font=("Helvetica", 15),
        )
        self.intructions_label_3 = Label(
            self.instructions_frm,
            text="3. Click <SAVE DATA> when all the data are filled in.",
            font=("Helvetica", 15),
        )
        self.instructions_frm.grid(row=0, column=0, sticky=W)
        self.intructions_label_1.grid(row=0, column=0, sticky=W)
        self.intructions_label_2.grid(row=1, column=0, sticky=W)
        self.intructions_label_3.grid(row=2, column=0, sticky=W)
        self.execute_frm = LabelFrame(
            self.main_frm, text="EXECUTE", font=Formats.LABELFRAME_HEADER_FORMAT.value
        )
        self.save_data_btn = Button(
            self.execute_frm,
            text="SAVE DATA",
            fg="green",
            command=lambda: self.__save_data(),
        )
        self.execute_frm.grid(row=1, column=0, sticky=W)
        self.save_data_btn.grid(row=0, column=0, sticky=W)
        self.video_frm = LabelFrame(
            self.main_frm,
            text="PROJECT VIDEOS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.column_names = [
            "INDEX",
            "VIDEO",
            "FPS",
            "RESOLUTION WIDTH",
            "RESOLUTION HEIGHT",
            "DISTANCE IN MM",
            "FIND DISTANCE",
            "PIXELS PER MM",
        ]
        self.col_widths = [
            "5",
            self.max_char_vid_name + 5,
            "18",
            "18",
            "18",
            "18",
            "18",
            "18",
        ]
        self.video_frm.grid(row=6, column=0)
        for cnt, (col_name, col_width) in enumerate(
            zip(self.column_names, self.col_widths)
        ):
            col_header_label = Label(
                self.video_frm,
                text=col_name,
                width=col_width,
                font=Formats.LABELFRAME_HEADER_FORMAT.value,
            )
            col_header_label.grid(row=0, column=cnt, sticky=W)
        self.videos = {}
        self.__append_videos_from_video_folder()
        self.__check_for_duplicate_names()
        self.duplicate_distance_btn = Button(
            self.video_frm,
            text="Duplicate index 1 distance/mm (CAUTION!)",
            fg="red",
            command=lambda: self.__duplicate_idx_1_distance_mm(),
        )
        self.duplicate_distance_btn.grid(row=1, column=5, sticky=W, padx=5)
        self.duplicate_btn = Button(
            self.video_frm,
            text="Duplicate index 1 pixels/mm (CAUTION!)",
            fg="red",
            command=lambda: self.__duplicate_idx_1_px_mm(),
        )
        self.duplicate_btn.grid(row=1, column=7, sticky=W, padx=5)

        for vid_cnt, name in enumerate(self.videos.keys()):
            vid_cnt += 2
            self.videos[name]["video_idx_lbl"].grid(
                row=vid_cnt, column=0, sticky=W, padx=5
            )
            self.videos[name]["video_name_lbl"].grid(
                row=vid_cnt, column=1, sticky=W, padx=5
            )
            self.videos[name]["fps_entry"].grid(row=vid_cnt, column=2, sticky=W)
            self.videos[name]["width_entry"].grid(
                row=vid_cnt, column=3, sticky=W, padx=5
            )
            self.videos[name]["height_entry"].grid(
                row=vid_cnt, column=4, sticky=W, padx=5
            )
            self.videos[name]["distance_entry"].grid(
                row=vid_cnt, column=5, sticky=W, padx=5
            )
            self.videos[name]["find_dist_btn"].grid(row=vid_cnt, column=6, padx=5)
            self.videos[name]["px_mm_entry"].grid(row=vid_cnt, column=7, padx=5)
        Label(
            self.execute_frm,
            text=f"THE PROJECT CONTAINS {len(self.videos.keys())} VIDEOS",
            fg="blue",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            padx=20,
        ).grid(row=0, column=1, sticky=NW)

    def __initiate_find_distance(self, k):
        video_name, distance = k[0], k[1].get()
        try:
            distance = float(distance)
        except:
            raise InvalidInputError(
                msg=f"The *DISTANCE IN MM* for video {video_name} is not an integer or float value. The *DISTANCE IN MM* has to be a numerical value.",
                source=self.__class__.__name__,
            )
        if distance <= 0:
            raise InvalidInputError(
                msg=f"The *DISTANCE IN MM* for video {video_name} is <=0. The *DISTANCE IN MM* has to be a value above 0.",
                source=self.__class__.__name__,
            )
        video_pixels_per_mm = get_coordinates_nilsson(
            os.path.join(self.video_dir, video_name), distance
        )
        self.videos[video_name]["px_mm"].set(str(round(video_pixels_per_mm, 5)))

    def __duplicate_idx_1_px_mm(self):
        px_value = self.videos[list(self.videos.keys())[0]]["px_mm_entry"].get()
        for cnt, name in enumerate(self.video_paths):
            self.videos[name]["px_mm"].set(px_value)

    def __duplicate_idx_1_distance_mm(self):
        px_value = self.videos[list(self.videos.keys())[0]]["distance_entry"].get()
        for cnt, name in enumerate(self.video_paths):
            self.videos[name]["distance"].set(px_value)

    def __save_data(self):
        self.video_df = pd.DataFrame(
            columns=[
                "Video",
                "fps",
                "Resolution_width",
                "Resolution_height",
                "Distance_in_mm",
                "pixels/mm",
            ]
        )
        for name, data in self.videos.items():
            _, name, _ = get_fn_ext(name)
            lst = [
                name,
                data["fps_var"].get(),
                data["width_var"].get(),
                data["height_var"].get(),
                data["distance"].get(),
                data["px_mm"].get(),
            ]
            self.video_df.loc[len(self.video_df)] = lst
        self.__check_no_zero_px_per_mm_values()
        self.video_df.drop_duplicates(subset=["Video"], inplace=True)
        self.video_df = self.video_df.set_index("Video")
        try:
            self.video_df.to_csv(
                os.path.join(self.project_path, Paths.VIDEO_INFO.value)
            )
        except PermissionError:
            raise PermissionError(
                msg="SimBA tried to write to project_folder/logs/video_info.csv, but was not allowed. If this file is open in another program, try closing it.",
                source=self.__class__.__name__,
            )
        stdout_success(
            msg="Video info saved at project_folder/logs/video_info.csv",
            source=self.__class__.__name__,
        )


# test = VideoInfoTable(config_path='/Users/simon/Desktop/envs/troubleshooting/locomotion/project_folder/project_config.ini')
# test.create_window()
# test.main_frm.mainloop()

# test = VideoInfoTable(config_path='/Users/simon/Desktop/troubleshooting/train_model_project/project_folder/project_config.ini')
# test.create_window()
# test.main_frm.mainloop()

#
# test = VideoInfoTable(config_path='/Users/simon/Desktop/envs/troubleshooting/sleap_5_animals_2/project_folder/project_config.ini')
# test.create_window()
# test.main_frm.mainloop()

# test = VideoInfoTable(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# test.create_window()
# test.main_frm.mainloop()



__author__ = "Simon Nilsson"

import multiprocessing
import os
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.ROI_plotter import ROIPlot
from simba.plotting.ROI_plotter_mp import ROIPlotMultiprocess
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, FileSelect)
from simba.utils.checks import check_float, check_if_filepath_list_is_empty
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.printing import stdout_success
from simba.utils.read_write import (find_files_of_filetypes_in_directory,
                                    get_fn_ext)


class VisualizeROITrackingPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="VISUALIZE ROI TRACKING", size=(800, 500))
        ConfigReader.__init__(self, config_path=config_path)
        self.video_list = []
        video_file_paths = find_files_of_filetypes_in_directory(
            directory=self.video_dir, extensions=[".mp4", ".avi"]
        )

        for file_path in video_file_paths:
            _, video_name, ext = get_fn_ext(filepath=file_path)
            self.video_list.append(video_name + ext)

        check_if_filepath_list_is_empty(
            filepaths=self.video_list,
            error_msg="No videos in SimBA project. Import videos into you SimBA project to visualize ROI tracking.",
        )
        self.multiprocess_var = BooleanVar()
        self.show_pose_var = BooleanVar(value=True)
        self.animal_name_var = BooleanVar(value=True)

        self.settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.ROI_DATA_PLOT.value,
        )
        self.threshold_entry_box = Entry_Box(
            self.settings_frm, "Body-part probability threshold", "30"
        )
        self.threshold_entry_box.entry_set(0.0)
        threshold_label = Label(
            self.settings_frm,
            text="Note: body-part locations detected with probabilities below this threshold is removed from visualization.",
            font=("Helvetica", 10, "italic"),
        )

        self.show_pose_cb = Checkbutton(
            self.settings_frm,
            text="Show pose-estimated location",
            variable=self.show_pose_var,
        )
        self.show_animal_name_cb = Checkbutton(
            self.settings_frm, text="Show animal names", variable=self.animal_name_var
        )
        self.multiprocess_cb = Checkbutton(
            self.settings_frm,
            text="Multi-process (faster)",
            variable=self.multiprocess_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.multiprocess_var,
                dropdown_menus=[self.multiprocess_dropdown],
            ),
        )
        self.multiprocess_dropdown = DropDownMenu(
            self.settings_frm, "CPU cores:", list(range(2, self.cpu_cnt)), "12"
        )
        self.multiprocess_dropdown.setChoices(2)
        self.multiprocess_dropdown.disable()

        self.body_parts_frm = LabelFrame(
            self.main_frm,
            text="SELECT BODY-PARTS",
            pady=10,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.animal_cnt_dropdown = DropDownMenu(
            self.body_parts_frm,
            "Number of animals",
            list(range(1, self.animal_cnt + 1)),
            "20",
            com=lambda x: self.__populate_bp_dropdown(bp_cnt=x),
        )
        self.animal_cnt_dropdown.setChoices(1)
        self.__populate_bp_dropdown(bp_cnt=1)

        self.run_frm = LabelFrame(
            self.main_frm,
            text="RUN VISUALIZATION",
            pady=10,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.single_video_frm = LabelFrame(
            self.run_frm,
            text="SINGLE video",
            pady=10,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.single_video_dropdown = DropDownMenu(
            self.single_video_frm,
            "Select video",
            self.video_list,
            "15",
            com=lambda x: self.update_file_select_box_from_dropdown(
                filename=x, fileselectbox=self.select_video_file_select
            ),
        )
        self.single_video_dropdown.setChoices(self.video_list[0])
        self.select_video_file_select = FileSelect(
            self.single_video_frm,
            "",
            lblwidth="1",
            file_types=[("VIDEO FILE", Options.ALL_VIDEO_FORMAT_STR_OPTIONS.value)],
            dropdown=self.single_video_dropdown,
        )
        self.select_video_file_select.filePath.set(self.video_list[0])
        self.single_video_btn = Button(
            self.single_video_frm,
            text="Create SINGLE ROI video",
            fg="blue",
            command=lambda: self.run_visualize(multiple=False),
        )

        self.all_videos_frm = LabelFrame(
            self.run_frm,
            text="ALL videos",
            pady=10,
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            fg="black",
        )
        self.all_videos_btn = Button(
            self.all_videos_frm,
            text="Create ALL ROI videos ({} videos found)".format(
                str(len(self.video_list))
            ),
            fg="red",
            command=lambda: self.run_visualize(multiple=True),
        )

        self.settings_frm.grid(row=0, column=0, sticky=NW)
        self.threshold_entry_box.grid(row=0, column=0, sticky=NW)
        threshold_label.grid(row=1, column=0, sticky=NW)
        self.show_pose_cb.grid(row=2, column=0, sticky=NW)
        self.show_animal_name_cb.grid(row=3, column=0, sticky=NW)
        self.multiprocess_cb.grid(row=4, column=0, sticky=NW)
        self.multiprocess_dropdown.grid(row=4, column=1, sticky=NW)

        self.body_parts_frm.grid(row=1, column=0, sticky=NW)
        self.animal_cnt_dropdown.grid(row=0, column=0, sticky=NW)

        self.run_frm.grid(row=2, column=0, sticky=NW)
        self.single_video_frm.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=0, sticky=NW)
        self.select_video_file_select.grid(row=0, column=1, sticky=NW)
        self.single_video_btn.grid(row=1, column=0, sticky=NW)
        self.all_videos_frm.grid(row=1, column=0, sticky=NW)
        self.all_videos_btn.grid(row=0, column=0, sticky=NW)
        self.main_frm.mainloop()

    def __populate_bp_dropdown(self, bp_cnt: int):
        if hasattr(self, "bp_dropdown_dict"):
            if len(self.bp_dropdown_dict.keys()) != bp_cnt:
                for k, v in self.bp_dropdown_dict.items():
                    v.destroy()

        self.bp_dropdown_dict = {}
        for cnt in range(int(self.animal_cnt_dropdown.getChoices())):
            self.bp_dropdown_dict[cnt] = DropDownMenu(
                self.body_parts_frm,
                self.multi_animal_id_list[cnt],
                self.body_parts_lst,
                "20",
            )
            self.bp_dropdown_dict[cnt].setChoices(self.body_parts_lst[cnt])
            self.bp_dropdown_dict[cnt].grid(row=cnt + 1, column=0, sticky=NW)

    def run_visualize(self, multiple: bool):
        if multiple:
            videos = self.video_list
        else:
            videos = [self.single_video_dropdown.getChoices()]

        check_float(
            name="Body-part probability threshold",
            value=self.threshold_entry_box.entry_get,
            min_value=0.0,
            max_value=1.0,
        )
        style_attr = {}
        style_attr["Show_body_part"] = False
        style_attr["Show_animal_name"] = False
        if self.show_pose_var.get():
            style_attr["Show_body_part"] = True
        if self.animal_name_var.get():
            style_attr["Show_animal_name"] = True

        body_parts = {}
        for k, v in self.bp_dropdown_dict.items():
            body_parts[self.multi_animal_id_list[k]] = v.getChoices()

        if not self.multiprocess_var.get():
            selected_video_path = os.path.join(
                self.video_dir, self.single_video_dropdown.getChoices()
            )
            self.check_if_selected_video_path_exist_in_project(
                video_path=selected_video_path
            )
            self.config.set(
                "ROI settings",
                "probability_threshold",
                str(self.threshold_entry_box.entry_get),
            )
            with open(self.config_path, "w") as f:
                self.config.write(f)
            for video in videos:
                roi_plotter = ROIPlot(
                    ini_path=self.config_path,
                    video_path=video,
                    style_attr=style_attr,
                    threshold=float(self.threshold_entry_box.entry_get),
                    body_parts=body_parts,
                )
                roi_plotter.insert_data()
                roi_plotter_multiprocessor = multiprocessing.Process(
                    target=roi_plotter.run()
                )
                roi_plotter_multiprocessor.start()

        else:
            with open(self.config_path, "w") as f:
                self.config.write(f)
            core_cnt = self.multiprocess_dropdown.getChoices()
            for video in videos:
                roi_plotter = ROIPlotMultiprocess(
                    ini_path=self.config_path,
                    video_path=video,
                    core_cnt=int(core_cnt),
                    style_attr=style_attr,
                    threshold=float(self.threshold_entry_box.entry_get),
                    body_parts=body_parts,
                )
                roi_plotter.run()

        stdout_success(
            msg="All ROI videos created and saved in project_folder/frames/output/ROI_analysis directory",
            source=self.__class__.__name__,
        )


# _ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Termites_5/project_folder/project_config.ini')
# _ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')
# _ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/simba_dev/tests/data/test_projects/mouse_open_field/project_folder/project_config.ini')
# _ = VisualizeROITrackingPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/dorian_2/project_folder/project_config.ini')



from __future__ import division

import glob
import math
import os

import numpy as np
import pandas as pd
from numba import jit

from simba.drop_bp_cords import *
from simba.drop_bp_cords import get_fn_ext, getBpHeaders, getBpNames
from simba.feature_extractors.unit_tests import (check_minimum_roll_windows,
                                                 read_video_info)
from simba.misc_tools import SimbaTimer
from simba.read_config_unit_tests import (check_if_filepath_list_is_empty,
                                          read_project_path_and_file_type)

TAIL_BP_NAMES = [
    "Zebrafish_Tail1",
    "Zebrafish_Tail2",
    "Zebrafish_Tail3",
    "Zebrafish_Tail4",
]
CENTER_BP_NAMES = ["Zebrafish_SwimBladder"]


class FishFeatureExtractor:
    def __init__(self, config_path: str):
        self.timer = SimbaTimer()
        self.timer.start_timer()
        self.windows_angular_dispersion_seconds = [10]
        self.compass_brackets = ["N", "NE", "E", "SE", "S", "SW", "W", "NW", "N"]
        self.compass_brackets_long = [
            "Direction_N",
            "Direction_NE",
            "Direction_E",
            "Direction_SE",
            "Direction_S",
            "Direction_SW",
            "Direction_W",
            "Direction_NW",
        ]
        self.compass_brackets_digits = ["0", "1", "2", "3", "4", "5", "6", "7", "0"]
        self.config = read_config_file(ini_path=config_path)
        self.project_path, self.file_type = read_project_path_and_file_type(
            config=self.config
        )
        self.input_file_dir = os.path.join(
            self.project_path, Paths.OUTLIER_CORRECTED.value
        )
        self.save_dir = os.path.join(
            self.project_path, Paths.FEATURES_EXTRACTED_DIR.value
        )
        self.video_info_path = os.path.join(self.project_path, Paths.VIDEO_INFO.value)
        self.video_info_df = pd.read_csv(self.video_info_path)
        bp_names_path = os.path.join(self.project_path, Paths.BP_NAMES.value)
        bp_names_df = pd.read_csv(bp_names_path, header=None)
        self.bp_names_list = list(bp_names_df[0])
        self.roll_windows_values = check_minimum_roll_windows(
            [10, 4, 2, 1, 0.1], self.video_info_df["fps"].min()
        )
        self.files_found = glob.glob(
            self.input_file_dir + "/*.{}".format(self.file_type)
        )
        print(
            "Extracting features from {} {}...".format(
                str(len(self.files_found)), "file(s)"
            )
        )
        check_if_filepath_list_is_empty(
            filepaths=self.files_found,
            error_msg="SIMBA ERROR: No file in {} directory".format(
                self.input_file_dir
            ),
        )

        for file_path in self.files_found:
            video_timer = SimbaTimer()
            video_timer.start_timer()
            self.roll_windows, self.angular_dispersion_windows = [], []
            dir_name, file_name, ext = get_fn_ext(file_path)
            self.save_path = os.path.join(self.save_dir, os.path.basename(file_path))
            currVideoSettings, self.currPixPerMM, self.fps = read_video_info(
                self.video_info_df, file_name
            )

            for i in range(len(self.roll_windows_values)):
                self.roll_windows.append(int(self.fps / self.roll_windows_values[i]))
            for i in range(len(self.windows_angular_dispersion_seconds)):
                self.angular_dispersion_windows.append(
                    int(self.fps * self.windows_angular_dispersion_seconds[i])
                )

            self.x_cols, self.y_cols, self.p_cols = getBpNames(str(config_path))
            self.x_cols_shifted, self.y_cols_shifted = [], []
            for col in zip(self.x_cols, self.y_cols, self.p_cols):
                self.x_cols_shifted.append(col[0] + "_shifted")
                self.y_cols_shifted.append(col[1] + "_shifted")

            bp_header_list = getBpHeaders(str(config_path))
            self.col_headers_shifted = []
            for bp in self.bp_names_list:
                self.col_headers_shifted.extend(
                    (bp + "_x_shifted", bp + "_y_shifted", bp + "_p_shifted")
                )

            csv_df = read_df(file_path, self.file_type)
            csv_df.columns = bp_header_list
            csv_df = csv_df.fillna(0).apply(pd.to_numeric)

            csv_df_shifted = csv_df.shift(periods=1)
            csv_df_shifted.columns = self.col_headers_shifted
            self.csv_df_combined = pd.concat(
                [csv_df, csv_df_shifted], axis=1, join="inner"
            ).fillna(0)
            self.calc_X_relative_to_Y_movement()
            self.calc_tail_and_center_movement()
            self.calc_X_relative_to_Y_movement_rolling_windows()
            self.calc_center_velocity()
            self.calc_rotation()
            self.calc_direction_switches()
            self.hot_end_encode_compass()
            self.calc_directional_switches_in_rolling_windows()
            self.calc_angular_dispersion()
            self.save_file()
            video_timer.stop_timer()
            print(
                f"Features extracted for video {file_name} (elapsed time {video_timer.elapsed_time_str}s)"
            )

        self.timer.stop_timer()
        print(
            f"Features extracted for all {str(len(self.files_found))} files, data saved in project_folder/csv/features_extracted directory (elapsed time {self.timer.elapsed_time_str}s)"
        )

    def angle2pt_degrees(self, ax, ay, bx, by):
        angle_degrees = math.degrees(math.atan2(ax - bx, by - ay))
        return angle_degrees + 360 if angle_degrees < 0 else angle_degrees

    def angle2pt_radians(self, degrees):
        angle_radians = degrees * math.pi / 180
        return angle_radians

    def angle2pt_sin(self, angle_radians):
        angle_sin = math.sin(angle_radians)
        return angle_sin

    def angle2pt_cos(self, angle_radians):
        angle_cos = math.cos(angle_radians)
        return angle_cos

    @staticmethod
    @jit(nopython=True)
    def euclidian_distance_calc(bp1xVals, bp1yVals, bp2xVals, bp2yVals):
        return np.sqrt((bp1xVals - bp2xVals) ** 2 + (bp1yVals - bp2yVals) ** 2)

    @staticmethod
    @jit(nopython=True)
    def angular_dispersion(cumsum_cos_np, cumsum_sin_np):
        out_array = np.empty((cumsum_cos_np.shape))
        for index in range(cumsum_cos_np.shape[0]):
            X, Y = cumsum_cos_np[index] / (index + 1), cumsum_sin_np[index] / (
                index + 1
            )
            out_array[index] = math.sqrt(X**2 + Y**2)
        return out_array

    def calc_angular_dispersion(self):
        dispersion_array = self.angular_dispersion(
            self.csv_df_combined["Angle_cos_cumsum"].values,
            self.csv_df_combined["Angle_sin_cumsum"].values,
        )
        self.csv_df_combined["Angular_dispersion"] = dispersion_array

        for win in range(len(self.angular_dispersion_windows)):
            col_name = "Angular_dispersion_window_" + str(
                self.windows_angular_dispersion_seconds[win]
            )
            self.csv_df_combined[col_name] = (
                self.csv_df_combined["Angular_dispersion"]
                .rolling(self.angular_dispersion_windows[win], min_periods=1)
                .mean()
            )

    def calc_X_relative_to_Y_movement(self):
        temp_df = pd.DataFrame()
        for bp in range(len(self.x_cols)):
            curr_x_col, curr_x_shifted_col, curr_y_col, curr_y_shifted_col = (
                self.x_cols[bp],
                self.x_cols_shifted[bp],
                self.y_cols[bp],
                self.y_cols_shifted[bp],
            )
            temp_df["x"] = (
                self.csv_df_combined[curr_x_col]
                - self.csv_df_combined[curr_x_shifted_col]
            )
            temp_df["y"] = (
                self.csv_df_combined[curr_y_col]
                - self.csv_df_combined[curr_y_shifted_col]
            )
            temp_df["Movement_{}_X_relative_2_Y".format(bp)] = (
                temp_df["x"] - temp_df["y"]
            )
            temp_df.drop(["x", "y"], axis=1, inplace=True)
        self.csv_df_combined["Movement_X_axis_relative_to_Y_axis"] = temp_df.sum(axis=1)

    def calc_tail_and_center_movement(self):
        tail_movement_col_names = []
        for tail_bp in TAIL_BP_NAMES:
            x_name, y_name = tail_bp + "_x", tail_bp + "_y"
            x_name_sh, y_name_sh = tail_bp + "_x_shifted", tail_bp + "_y_shifted"
            self.csv_df_combined[tail_bp + "_movement"] = (
                self.euclidian_distance_calc(
                    self.csv_df_combined[x_name].values,
                    self.csv_df_combined[y_name].values,
                    self.csv_df_combined[x_name_sh].values,
                    self.csv_df_combined[y_name_sh].values,
                )
                / self.currPixPerMM
            )
            tail_movement_col_names.append(tail_bp + "_movement")
        self.csv_df_combined["total_tail_bp_movement"] = self.csv_df_combined[
            tail_movement_col_names
        ].sum(axis=1)

        for center_bp in CENTER_BP_NAMES:
            x_name, y_name = center_bp + "_x", center_bp + "_y"
            x_name_sh, y_name_sh = center_bp + "_x_shifted", center_bp + "_y_shifted"
            self.csv_df_combined[center_bp + "_movement"] = (
                self.euclidian_distance_calc(
                    self.csv_df_combined[x_name].values,
                    self.csv_df_combined[y_name].values,
                    self.csv_df_combined[x_name_sh].values,
                    self.csv_df_combined[y_name_sh].values,
                )
                / self.currPixPerMM
            )
            self.csv_df_combined[center_bp + "_cum_distance_travelled"] = (
                self.csv_df_combined[center_bp + "_movement"].cumsum()
            )

    def calc_X_relative_to_Y_movement_rolling_windows(self):
        for i in range(len(self.roll_windows_values)):
            currentColName = "Movement_X_axis_relative_to_Y_axis_mean_" + str(
                self.roll_windows_values[i]
            )
            self.csv_df_combined[currentColName] = (
                self.csv_df_combined["Movement_X_axis_relative_to_Y_axis"]
                .rolling(self.roll_windows[i], min_periods=1)
                .mean()
            )
            currentColName = "Movement_X_axis_relative_to_Y_axis_sum_" + str(
                self.roll_windows_values[i]
            )
            self.csv_df_combined[currentColName] = (
                self.csv_df_combined["Movement_X_axis_relative_to_Y_axis"]
                .rolling(self.roll_windows[i], min_periods=1)
                .sum()
            )

    def calc_directional_switches_in_rolling_windows(self):
        for win in range(len(self.roll_windows_values)):
            currentColName = "Number_of_direction_switches_" + str(
                self.roll_windows_values[win]
            )
            self.csv_df_combined[currentColName] = (
                self.csv_df_combined["Direction_switch"]
                .rolling(self.roll_windows[win], min_periods=1)
                .sum()
            )
            currentColName = "Directionality_of_switches_switches_" + str(
                self.roll_windows_values[win]
            )
            self.csv_df_combined[currentColName] = (
                self.csv_df_combined["Switch_direction_value"]
                .rolling(self.roll_windows[win], min_periods=1)
                .sum()
            )

    def calc_center_velocity(self):
        for center_bp in CENTER_BP_NAMES:
            self.csv_df_combined[center_bp + "_velocity"] = (
                self.csv_df_combined[center_bp + "_movement"]
                .rolling(int(self.fps), min_periods=1)
                .sum()
            )

    def calc_rotation(self):
        self.csv_df_combined["Clockwise_angle_degrees"] = self.csv_df_combined.apply(
            lambda x: self.angle2pt_degrees(
                x[CENTER_BP_NAMES[0] + "_x"],
                x[CENTER_BP_NAMES[0] + "_y"],
                x[TAIL_BP_NAMES[0] + "_x"],
                x[TAIL_BP_NAMES[0] + "_y"],
            ),
            axis=1,
        )
        self.csv_df_combined["Angle_radians"] = self.angle2pt_radians(
            self.csv_df_combined["Clockwise_angle_degrees"]
        )
        self.csv_df_combined["Angle_sin"] = self.csv_df_combined.apply(
            lambda x: self.angle2pt_sin(x["Angle_radians"]), axis=1
        )
        self.csv_df_combined["Angle_cos"] = self.csv_df_combined.apply(
            lambda x: self.angle2pt_cos(x["Angle_radians"]), axis=1
        )
        self.csv_df_combined["Angle_sin_cumsum"] = self.csv_df_combined[
            "Angle_sin"
        ].cumsum()
        self.csv_df_combined["Angle_cos_cumsum"] = self.csv_df_combined[
            "Angle_cos"
        ].cumsum()
        compass_lookup = list(
            round(self.csv_df_combined["Clockwise_angle_degrees"] / 45)
        )
        compass_lookup = [int(i) for i in compass_lookup]
        compasFaceList_bracket, compasFaceList_digit = [], []
        for compasDirection in compass_lookup:
            compasFaceList_bracket.append(self.compass_brackets[compasDirection])
            compasFaceList_digit.append(self.compass_brackets_digits[compasDirection])
        self.csv_df_combined["Compass_direction"] = compasFaceList_bracket
        self.csv_df_combined["Compass_digit"] = compasFaceList_digit

        for i in range(len(self.roll_windows_values)):
            column_name = "Mean_angle_time_window_{}".format(
                str(self.roll_windows_values[i])
            )
            self.csv_df_combined[column_name] = (
                self.csv_df_combined["Clockwise_angle_degrees"]
                .rolling(self.roll_windows[i], min_periods=1)
                .mean()
            )

    def hot_end_encode_compass(self):
        compass_hot_end = pd.get_dummies(
            self.csv_df_combined["Compass_direction"], prefix="Direction"
        )
        compass_hot_end = compass_hot_end.T.reindex(
            self.compass_brackets_long
        ).T.fillna(0)
        self.csv_df_combined = pd.concat(
            [self.csv_df_combined, compass_hot_end], axis=1
        )

    def calc_direction_switches(self):
        self.grouped_df = pd.DataFrame()
        v = (
            self.csv_df_combined["Compass_digit"]
            != self.csv_df_combined["Compass_digit"].shift()
        ).cumsum()
        u = self.csv_df_combined.groupby(v)["Compass_digit"].agg(["all", "count"])
        m = u["all"] & u["count"].ge(1)
        self.grouped_df["groups"] = self.csv_df_combined.groupby(v).apply(
            lambda x: (x.index[0], x.index[-1])
        )[m]
        currdirectionList, DirectionSwitchIndexList, currdirectionListValue = [], [], []
        for indexes, row in self.grouped_df.iterrows():
            currdirectionList.append(
                self.csv_df_combined.loc[row["groups"][0]]["Compass_direction"]
            )
            DirectionSwitchIndexList.append(row["groups"][1])
            currdirectionListValue.append(
                self.csv_df_combined.loc[row["groups"][0]]["Compass_digit"]
            )
        self.grouped_df["Direction_switch"] = currdirectionList
        self.grouped_df["Direction_value"] = currdirectionListValue
        self.csv_df_combined.loc[DirectionSwitchIndexList, "Direction_switch"] = 1
        self.csv_df_combined["Compass_digit_shifted"] = self.csv_df_combined[
            "Compass_digit"
        ].shift(-1)
        self.csv_df_combined = self.csv_df_combined.fillna(0)
        self.csv_df_combined["Switch_direction_value"] = self.csv_df_combined.apply(
            lambda x: self.calc_switch_direction(
                x["Compass_digit_shifted"], x["Compass_digit"]
            ),
            axis=1,
        )

    def calc_switch_direction(self, compass_digit_shifted, compass_digit):
        if (compass_digit_shifted == "0") and (compass_digit == "7"):
            return 1
        else:
            return int(compass_digit_shifted) - int(compass_digit)

    def save_file(self):
        self.csv_df_combined = self.csv_df_combined.drop(
            self.col_headers_shifted, axis=1
        )
        self.csv_df_combined = self.csv_df_combined.drop(
            [
                "Compass_digit_shifted",
                "Direction_switch",
                "Switch_direction_value",
                "Compass_digit",
                "Compass_direction",
                "Angle_sin_cumsum",
                "Angle_cos_cumsum",
            ],
            axis=1,
        )
        save_df(self.csv_df_combined, self.file_type, self.save_path)



__author__ = "Simon Nilsson"

import os
from collections import defaultdict
from copy import deepcopy
from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.path_plotter import PathPlotterSingleCore
from simba.plotting.path_plotter_mp import PathPlotterMulticore
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box)
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_if_string_value_is_valid_video_timestamp,
                                check_if_valid_rgb_str, check_int,
                                check_that_hhmmss_start_is_before_end)
from simba.utils.enums import Formats, Keys, Links, Paths
from simba.utils.errors import FrameRangeError
from simba.utils.read_write import get_file_name_info_in_directory


class PathPlotPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        ConfigReader.__init__(self, config_path=config_path)
        self.data_path = os.path.join(
            self.project_path, Paths.MACHINE_RESULTS_DIR.value
        )
        self.files_found_dict = get_file_name_info_in_directory(
            directory=self.data_path, file_type=self.file_type
        )
        check_if_filepath_list_is_empty(
            filepaths=list(self.files_found_dict.keys()),
            error_msg="SIMBA ERROR: Zero files found in the project_folder/csv/machine_results directory. Create classification results before visualizing path plots",
        )

        PopUpMixin.__init__(self, title="CREATE PATH PLOTS", size=(550, 850))
        self.resolution_options = deepcopy(self.resolutions)
        self.resolution_options.insert(0, "As input")
        self.bg_clr_options = deepcopy(list(self.colors_dict.keys()))
        self.animal_trace_clrs = deepcopy(list(self.colors_dict.keys()))
        self.bg_clr_options.extend(("Video - static frame", "Video - moving frames"))
        self.animal_trace_clrs.append("Custom")
        self.bg_opacity_options = [str(x) + "%" for x in range(10, 110, 10)]
        self.animal_cnt_options = list(range(1, self.animal_cnt + 1))
        self.custom_rgb_selections = {}

        self.style_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="STYLE SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.PATH_PLOTS.value,
        )
        self.autocompute_var = BooleanVar(value=True)
        self.auto_compute_styles = Checkbutton(
            self.style_settings_frm,
            text="Auto-compute styles",
            variable=self.autocompute_var,
            command=self.enable_style_settings,
        )
        self.max_prior_lines_dropdown = DropDownMenu(
            self.style_settings_frm,
            "Max prior lines:",
            ["Entire video", "Specify milliseconds"],
            "16",
            com=self.enable_entrybox_from_dropdown,
        )
        self.max_lines_entry = Entry_Box(
            self.style_settings_frm,
            "Max prior lines (ms): ",
            "16",
            validation="numeric",
        )
        self.resolution_dropdown = DropDownMenu(
            self.style_settings_frm, "Resolution:", self.resolution_options, "16"
        )
        self.bg_clr_dropdown = DropDownMenu(
            self.style_settings_frm,
            "Background:",
            self.bg_clr_options,
            "16",
            com=lambda x: self.__activate_settings(choice=x),
        )
        self.line_width = Entry_Box(
            self.style_settings_frm, "Line width: ", "16", validation="numeric"
        )
        self.font_size = Entry_Box(
            self.style_settings_frm, "Font size: ", "16", validation="numeric"
        )
        self.font_thickness = Entry_Box(
            self.style_settings_frm, "Font thickness: ", "16", validation="numeric"
        )
        self.bg_opacity_dropdown = DropDownMenu(
            self.style_settings_frm,
            "Background opacity:",
            self.bg_opacity_options,
            "16",
        )
        self.circle_size = Entry_Box(
            self.style_settings_frm, "Circle size: ", "16", validation="numeric"
        )
        self.resolution_dropdown.setChoices(self.resolution_options[0])
        self.line_width.entry_set(val=6)
        self.bg_clr_dropdown.setChoices("White")
        self.circle_size.entry_set(val=20)
        self.font_size.entry_set(val=3)
        self.font_thickness.entry_set(val=2)
        self.max_prior_lines_dropdown.setChoices("Entire video")
        self.max_prior_lines_dropdown.disable()
        self.max_lines_entry.entry_set(2000)
        self.bg_opacity_dropdown.setChoices("100%")
        self.bg_opacity_dropdown.disable()
        self.resolution_dropdown.disable()
        self.line_width.set_state("disable")
        self.bg_clr_dropdown.disable()
        self.circle_size.set_state("disable")
        self.font_size.set_state("disable")
        self.max_lines_entry.set_state("disable")
        self.font_thickness.set_state("disable")

        self.body_parts_frm = LabelFrame(
            self.main_frm,
            text="CHOOSE BODY-PARTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.number_of_animals_dropdown = DropDownMenu(
            self.body_parts_frm,
            "# Animals:",
            self.animal_cnt_options,
            "16",
            com=self.populate_body_parts_menu,
        )
        self.number_of_animals_dropdown.setChoices(self.animal_cnt_options[0])

        self.video_slicing_frm = LabelFrame(
            self.main_frm,
            text="SEGMENTS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.slice_var = BooleanVar(value=False)
        self.video_start_time_entry = Entry_Box(
            self.video_slicing_frm, "Start time:", "16"
        )
        self.video_end_time_entry = Entry_Box(self.video_slicing_frm, "End time:", "16")
        self.video_start_time_entry.entry_set("00:00:00")
        self.video_end_time_entry.entry_set("00:00:00")
        self.video_start_time_entry.set_state("disable")
        self.video_end_time_entry.set_state("disable")
        self.slice_cb = Checkbutton(
            self.video_slicing_frm,
            text="Plot only defined time-segment",
            variable=self.slice_var,
            command=lambda: self.enable_entrybox_from_checkbox(
                check_box_var=self.slice_var,
                entry_boxes=[self.video_start_time_entry, self.video_end_time_entry],
            ),
        )
        self.slice_cb.grid(row=0, column=0, sticky=NW)
        self.video_start_time_entry.grid(row=1, column=0, sticky=NW)
        self.video_end_time_entry.grid(row=2, column=0, sticky=NW)

        self.clf_frm = LabelFrame(
            self.main_frm,
            text="CHOOSE CLASSIFICATION VISUALIZATION",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.include_clf_locations_var = BooleanVar(value=False)
        self.include_clf_locations_cb = Checkbutton(
            self.clf_frm,
            text="Include classification locations",
            variable=self.include_clf_locations_var,
            command=self.populate_clf_location_data,
        )
        self.include_clf_locations_cb.grid(row=0, sticky=NW)
        self.populate_clf_location_data()

        self.populate_body_parts_menu(self.animal_cnt_options[0])
        self.settings_frm = LabelFrame(
            self.main_frm,
            text="VISUALIZATION SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
        )
        self.path_frames_var = BooleanVar()
        self.path_videos_var = BooleanVar()
        self.path_last_frm_var = BooleanVar()
        self.multiprocessing_var = BooleanVar()
        self.include_animal_names_var = BooleanVar(value=True)

        path_frames_cb = Checkbutton(
            self.settings_frm, text="Create frames", variable=self.path_frames_var
        )
        path_videos_cb = Checkbutton(
            self.settings_frm, text="Create videos", variable=self.path_videos_var
        )
        path_last_frm_cb = Checkbutton(
            self.settings_frm, text="Create last frame", variable=self.path_last_frm_var
        )
        self.include_animal_names_cb = Checkbutton(
            self.settings_frm,
            text="Include animal names",
            variable=self.include_animal_names_var,
        )

        self.multiprocess_cb = Checkbutton(
            self.settings_frm,
            text="Multiprocess videos (faster)",
            variable=self.multiprocessing_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.multiprocessing_var,
                dropdown_menus=[self.multiprocess_dropdown],
            ),
        )
        self.multiprocess_dropdown = DropDownMenu(
            self.settings_frm, "CPU cores:", list(range(2, self.cpu_cnt)), "12"
        )
        self.multiprocess_dropdown.setChoices(2)
        self.multiprocess_dropdown.disable()

        self.run_frm = LabelFrame(
            self.main_frm,
            text="RUN",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_frm = LabelFrame(
            self.run_frm,
            text="SINGLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_single_video_btn = Button(
            self.run_single_video_frm,
            text="Create single video",
            fg="blue",
            command=lambda: self.__create_path_plots(multiple_videos=False),
        )
        self.single_video_dropdown = DropDownMenu(
            self.run_single_video_frm,
            "Video:",
            list(self.files_found_dict.keys()),
            "12",
        )
        self.single_video_dropdown.setChoices(list(self.files_found_dict.keys())[0])

        self.run_multiple_videos = LabelFrame(
            self.run_frm,
            text="MULTIPLE VIDEO",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
            pady=5,
            padx=5,
            fg="black",
        )
        self.run_multiple_video_btn = Button(
            self.run_multiple_videos,
            text="Create multiple videos ({} video(s) found)".format(
                str(len(list(self.files_found_dict.keys())))
            ),
            fg="blue",
            command=lambda: self.__create_path_plots(multiple_videos=True),
        )

        self.style_settings_frm.grid(row=0, sticky=NW)
        self.auto_compute_styles.grid(row=0, sticky=NW)
        self.resolution_dropdown.grid(row=1, sticky=NW)
        self.max_prior_lines_dropdown.grid(row=2, sticky=NW)
        self.max_lines_entry.grid(row=3, sticky=NW)
        self.line_width.grid(row=4, sticky=NW)
        self.circle_size.grid(row=5, sticky=NW)
        self.font_size.grid(row=6, sticky=NW)
        self.font_thickness.grid(row=7, sticky=NW)
        self.bg_clr_dropdown.grid(row=8, sticky=NW)
        self.bg_opacity_dropdown.grid(row=9, sticky=NW)

        self.video_slicing_frm.grid(row=1, sticky=NW)
        self.clf_frm.grid(row=2, sticky=NW)

        self.body_parts_frm.grid(row=3, sticky=NW)
        self.number_of_animals_dropdown.grid(row=0, sticky=NW)

        self.settings_frm.grid(row=4, sticky=NW)
        path_frames_cb.grid(row=0, sticky=NW)
        path_videos_cb.grid(row=1, sticky=NW)
        path_last_frm_cb.grid(row=2, sticky=NW)
        self.include_animal_names_cb.grid(row=3, sticky=NW)
        self.multiprocess_cb.grid(row=4, column=0, sticky=NW)
        self.multiprocess_dropdown.grid(row=4, column=1, sticky=NW)

        self.run_frm.grid(row=5, sticky=NW)
        self.run_single_video_frm.grid(row=0, sticky=NW)
        self.run_single_video_btn.grid(row=0, column=0, sticky=NW)
        self.single_video_dropdown.grid(row=0, column=1, sticky=NW)

        self.run_multiple_videos.grid(row=1, sticky=NW)
        self.run_multiple_video_btn.grid(row=0, sticky=NW)

        self.main_frm.mainloop()

    def populate_body_parts_menu(self, choice):
        if hasattr(self, "bp_dropdowns"):
            for (k, v), (k2, v2) in zip(
                self.bp_dropdowns.items(), self.bp_colors.items()
            ):
                self.bp_dropdowns[k].destroy()
                self.bp_colors[k].destroy()
        for k, v in self.custom_rgb_selections.items():
            v.destroy()
        self.custom_rgb_selections = {}

        self.bp_dropdowns, self.bp_colors = {}, {}
        self.bp_row_idx = []
        for animal_cnt in range(int(self.number_of_animals_dropdown.getChoices())):
            self.bp_dropdowns[animal_cnt] = DropDownMenu(
                self.body_parts_frm,
                "Body-part {}:".format(str(animal_cnt + 1)),
                self.body_parts_lst,
                "16",
            )
            self.bp_dropdowns[animal_cnt].setChoices(self.body_parts_lst[animal_cnt])
            self.bp_dropdowns[animal_cnt].grid(row=animal_cnt + 1, column=0, sticky=NW)
            self.bp_colors[animal_cnt] = DropDownMenu(
                self.body_parts_frm,
                "",
                self.animal_trace_clrs,
                "2",
                com=lambda x, k=animal_cnt: self.__set_custom_clrs(choice=x, row=k),
            )
            self.bp_colors[animal_cnt].setChoices(
                list(self.colors_dict.keys())[animal_cnt]
            )
            self.bp_colors[animal_cnt].grid(row=animal_cnt + 1, column=1, sticky=NW)

    def __activate_settings(self, choice: str):
        if choice == "Video - static frame" or choice == "Video - moving frames":
            self.bg_opacity_dropdown.enable()
        else:
            self.bg_opacity_dropdown.disable()
        if choice == "Video - static frame":
            self.static_frm_index_eb = Entry_Box(
                self.style_settings_frm, "Frame index: ", "10", validation="numeric"
            )
            self.static_frm_index_eb.entry_set(val=1)
            self.static_frm_index_eb.grid(row=8, column=1, sticky=NW)
        else:
            if hasattr(self, "static_frm_index_eb"):
                self.static_frm_index_eb.destroy()

    def __set_custom_clrs(self, choice: str, row: int):
        if choice == "Custom":
            self.custom_rgb_selections[row] = Entry_Box(
                self.body_parts_frm, "RGB:", "5", entry_box_width=10
            )
            self.custom_rgb_selections[row].entry_set(val="255,0,0")
            self.custom_rgb_selections[row].grid(row=row + 1, column=3, sticky=NW)
        else:
            if row in self.custom_rgb_selections.keys():
                self.custom_rgb_selections[row].destroy()
                self.custom_rgb_selections.pop(row)

    def populate_clf_location_data(self):
        self.clf_name, self.clf_clr, self.clf_size = {}, {}, {}
        size_lst = list(range(1, 51))
        size_lst = ["Size: " + str(x) for x in size_lst]
        for clf_cnt, clf_name in enumerate(self.clf_names):
            self.clf_name[clf_cnt] = DropDownMenu(
                self.clf_frm,
                "Classifier {}:".format(str(clf_cnt + 1)),
                self.clf_names,
                "16",
            )
            self.clf_name[clf_cnt].setChoices(self.clf_names[clf_cnt])
            self.clf_name[clf_cnt].grid(row=clf_cnt + 1, column=0, sticky=NW)

            self.clf_clr[clf_cnt] = DropDownMenu(
                self.clf_frm, "", list(self.colors_dict.keys()), "2"
            )
            self.clf_clr[clf_cnt].setChoices(list(self.colors_dict.keys())[clf_cnt])
            self.clf_clr[clf_cnt].grid(row=clf_cnt + 1, column=1, sticky=NW)

            self.clf_size[clf_cnt] = DropDownMenu(self.clf_frm, "", size_lst, "2")
            self.clf_size[clf_cnt].setChoices(size_lst[15])
            self.clf_size[clf_cnt].grid(row=clf_cnt + 1, column=2, sticky=NW)

        self.enable_clf_location_settings()

    def enable_clf_location_settings(self):
        if self.include_clf_locations_var.get():
            for clf_cnt in self.clf_name.keys():
                self.clf_name[clf_cnt].enable()
                self.clf_clr[clf_cnt].enable()
                self.clf_size[clf_cnt].enable()
        else:
            for clf_cnt in self.clf_name.keys():
                self.clf_name[clf_cnt].disable()
                self.clf_clr[clf_cnt].disable()
                self.clf_size[clf_cnt].disable()

    def enable_entrybox_from_dropdown(self, dropdown_selection):
        if dropdown_selection == "Entire video":
            self.max_lines_entry.set_state("disable")
        else:
            self.max_lines_entry.set_state("normal")

    def enable_style_settings(self):
        if not self.autocompute_var.get():
            self.resolution_dropdown.enable()
            self.max_prior_lines_dropdown.enable()
            self.max_lines_entry.set_state("normal")
            self.line_width.set_state("normal")
            self.font_thickness.set_state("normal")
            self.circle_size.set_state("normal")
            self.font_size.set_state("normal")
            self.bg_clr_dropdown.enable()
            if self.bg_clr_dropdown.getChoices() == "Video":
                self.bg_opacity_dropdown.enable()
            self.enable_entrybox_from_dropdown(
                self.max_prior_lines_dropdown.getChoices()
            )
        else:
            self.resolution_dropdown.disable()
            self.max_prior_lines_dropdown.disable()
            self.max_lines_entry.set_state("disable")
            self.font_thickness.set_state("disable")
            self.line_width.set_state("disable")
            self.circle_size.set_state("disable")
            self.font_size.set_state("disable")
            self.bg_clr_dropdown.disable()
            self.bg_opacity_dropdown.disable()

    def __create_path_plots(self, multiple_videos: bool):
        if self.autocompute_var.get():
            style_attr = None
        else:
            if self.resolution_dropdown.getChoices() != "As input":
                width = int(self.resolution_dropdown.getChoices().split("×")[0])
                height = int(self.resolution_dropdown.getChoices().split("×")[1])
            else:
                width, height = "As input", "As input"
            check_int(
                name="PATH LINE WIDTH", value=self.line_width.entry_get, min_value=1
            )
            check_int(
                name="PATH CIRCLE SIZE", value=self.circle_size.entry_get, min_value=1
            )
            check_int(
                name="PATH FONT SIZE", value=self.font_size.entry_get, min_value=1
            )
            check_int(
                name="FONT THICKNESS", value=self.font_thickness.entry_get, min_value=1
            )
            if self.bg_clr_dropdown.getChoices() == "Video - static frame":
                check_int(
                    name="Static frame index",
                    value=self.static_frm_index_eb.entry_get,
                    min_value=0,
                )
                bg_clr = {
                    "type": "static",
                    "opacity": int(
                        "".join(
                            c
                            for c in self.bg_opacity_dropdown.getChoices()
                            if c.isdigit()
                        )
                    ),
                    "frame_index": int(self.static_frm_index_eb.entry_get),
                }
            elif self.bg_clr_dropdown.getChoices() == "Video - moving frames":
                bg_clr = {
                    "type": "moving",
                    "opacity": int(
                        "".join(
                            c
                            for c in self.bg_opacity_dropdown.getChoices()
                            if c.isdigit()
                        )
                    ),
                }
            else:
                bg_clr = self.bg_clr_dropdown.getChoices()

            style_attr = {
                "width": width,
                "height": height,
                "line width": int(self.line_width.entry_get),
                "font size": int(self.font_size.entry_get),
                "font thickness": int(self.font_thickness.entry_get),
                "circle size": int(self.circle_size.entry_get),
                "bg color": bg_clr,
                "clf locations": self.include_clf_locations_var.get(),
            }

            if self.max_prior_lines_dropdown.getChoices() == "Entire video":
                style_attr["max lines"] = "entire video"
            else:
                check_int(
                    name="PATH MAX LINES",
                    value=self.max_lines_entry.entry_get,
                    min_value=1,
                )
                style_attr["max lines"] = self.max_lines_entry.entry_get

        animal_attr = defaultdict(list)
        for attr in (self.bp_dropdowns, self.bp_colors):
            for key, value in attr.items():
                animal_attr[key].append(value.getChoices())
        for c, (k, v) in enumerate(animal_attr.items()):
            if v[1] == "Custom":
                custom_rgb_val = self.custom_rgb_selections[c].entry_get
                custom_rgb_val = check_if_valid_rgb_str(input=custom_rgb_val)
                animal_attr[k] = [v[0], custom_rgb_val]

        self.slicing = None
        if self.slice_var.get():
            check_if_string_value_is_valid_video_timestamp(
                value=self.video_start_time_entry.entry_get,
                name="Video slicing START TIME",
            )
            check_if_string_value_is_valid_video_timestamp(
                value=self.video_end_time_entry.entry_get, name="Video slicing END TIME"
            )
            if (
                self.video_start_time_entry.entry_get
                == self.video_end_time_entry.entry_get
            ):
                raise FrameRangeError(
                    msg="The sliced start and end times cannot be identical",
                    source=self.__class__.__name__,
                )
            check_that_hhmmss_start_is_before_end(
                start_time=self.video_start_time_entry.entry_get,
                end_time=self.video_end_time_entry.entry_get,
                name="SLICE TIME STAMPS",
            )
            self.slicing = {
                "start_time": self.video_start_time_entry.entry_get,
                "end_time": self.video_end_time_entry.entry_get,
            }

        clf_attr = None
        if self.include_clf_locations_var.get():
            clf_attr = defaultdict(list)
            for attr in (self.clf_name, self.clf_clr, self.clf_size):
                for key, value in attr.items():
                    clf_attr[key].append(value.getChoices())

        if multiple_videos:
            data_paths = list(self.files_found_dict.values())
        else:
            data_paths = [
                self.files_found_dict[self.single_video_dropdown.getChoices()]
            ]

        if not self.multiprocessing_var.get():
            path_plotter = PathPlotterSingleCore(
                config_path=self.config_path,
                frame_setting=self.path_frames_var.get(),
                video_setting=self.path_videos_var.get(),
                last_frame=self.path_last_frm_var.get(),
                files_found=data_paths,
                input_style_attr=style_attr,
                print_animal_names=self.include_animal_names_var.get(),
                animal_attr=animal_attr,
                input_clf_attr=clf_attr,
                slicing=self.slicing,
            )
        else:
            path_plotter = PathPlotterMulticore(
                config_path=self.config_path,
                frame_setting=self.path_frames_var.get(),
                video_setting=self.path_videos_var.get(),
                last_frame=self.path_last_frm_var.get(),
                files_found=data_paths,
                input_style_attr=style_attr,
                print_animal_names=self.include_animal_names_var.get(),
                animal_attr=animal_attr,
                input_clf_attr=clf_attr,
                cores=int(self.multiprocess_dropdown.getChoices()),
                slicing=self.slicing,
            )

        path_plotter.run()


# _ = PathPlotPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini')

# _ = PathPlotPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/dorian_2/project_folder/project_config.ini')



__author__ = "Simon Nilsson"


import os

import cv2
import numpy as np


def get_coordinates_nilsson(filenames, knownmm):
    global cordStatus
    global moveStatus
    global insertStatus
    global changeLoop
    cordStatus = False
    moveStatus = False
    insertStatus = False
    changeLoop = False

    newCordList = []

    # mouse callback function
    def draw_circle(event, x, y, flags, param):
        global ix, iy, cordStatus
        if (event == cv2.EVENT_LBUTTONDBLCLK) and len(cordList) < 4:
            cv2.circle(overlay, (x, y), circleScale, (144, 0, 255), -1)
            cordList.append(x)
            cordList.append(y)
            if len(cordList) == 4:
                cordStatus = True
                cv2.line(
                    overlay,
                    (cordList[0], cordList[1]),
                    (cordList[2], cordList[3]),
                    (144, 0, 255),
                    6,
                )

    def select_cord_to_change(event, x, y, flags, param):
        global moveStatus
        global coordChange
        if event == cv2.EVENT_LBUTTONDBLCLK:
            if (
                (x >= (cordList[0] - 20))
                and (x <= (cordList[0] + 20))
                and (y >= (cordList[1] - 20))
                and (y <= (cordList[1] + 20))
            ):  # change point1
                coordChange = [1, cordList[0], cordList[1]]
                moveStatus = True
            if (
                (x >= (cordList[2] - 20))
                and (x <= (cordList[2] + 20))
                and (y >= (cordList[3] - 20))
                and (y <= (cordList[3] + 20))
            ):  # change point2
                coordChange = [2, cordList[2], cordList[3]]
                moveStatus = True

    def select_new_dot_location(event, x, y, flags, param):
        global insertStatus
        if event == cv2.EVENT_LBUTTONDBLCLK:
            newCordList.append(x)
            newCordList.append(y)
            insertStatus = True

    # extract one frame
    currentDir = str(os.path.dirname(filenames))
    videoName = str(os.path.basename(filenames))

    os.chdir(currentDir)
    try:
        cap = cv2.VideoCapture(videoName)
        cap.set(1, 0)
        ret, frame = cap.read()
        fileName = str(0) + str(".bmp")
        filePath = os.path.join(currentDir, fileName)
        cv2.imwrite(filePath, frame)
        img = cv2.imread(filePath)
        (imageHeight, imageWidth) = img.shape[:2]
    except AttributeError:
        print(
            "ERROR: Make sure the video file "
            + str(videoName)
            + " is located in your project_folder/videos directory"
        )
    maxResDimension = max(imageWidth, imageHeight)
    mySpaceScale, myRadius, myResolution, myFontScale = 80, 20, 1500, 1.5
    circleScale = int(myRadius / (myResolution / maxResDimension))
    fontScale = float(myFontScale / (myResolution / maxResDimension))
    spacingScale = int(mySpaceScale / (myResolution / maxResDimension))
    origImage = img.copy()
    overlay = img.copy()
    ix, iy = -1, -1
    cordList = []
    cv2.namedWindow(
        "Select coordinates: double left mouse click at two locations. Press ESC when done",
        cv2.WINDOW_NORMAL,
    )

    while 1:
        if cordStatus == False and (moveStatus == False) and (insertStatus == False):
            cv2.setMouseCallback(
                "Select coordinates: double left mouse click at two locations. Press ESC when done",
                draw_circle,
            )
            cv2.imshow(
                "Select coordinates: double left mouse click at two locations. Press ESC when done",
                overlay,
            )
            k = cv2.waitKey(20) & 0xFF
            if k == 27:
                break
        if (cordStatus == True) and (moveStatus == False) and (insertStatus == False):
            if changeLoop == True:
                overlay = origImage.copy()
                cv2.circle(
                    overlay, (cordList[0], cordList[1]), circleScale, (144, 0, 255), -1
                )
                cv2.circle(
                    overlay, (cordList[2], cordList[3]), circleScale, (144, 0, 255), -1
                )
                cv2.line(
                    overlay,
                    (cordList[0], cordList[1]),
                    (cordList[2], cordList[3]),
                    (144, 0, 255),
                    int(circleScale / 5),
                )
            cv2.putText(
                overlay,
                "Click on circle to move",
                (20, 30),
                cv2.FONT_HERSHEY_TRIPLEX,
                fontScale,
                (255, 0, 255),
                2,
            )
            cv2.putText(
                overlay,
                "Press ESC to save and exit",
                (20, 50 + spacingScale),
                cv2.FONT_HERSHEY_TRIPLEX,
                fontScale,
                (255, 0, 255),
                2,
            )
            cv2.imshow(
                "Select coordinates: double left mouse click at two locations. Press ESC when done",
                overlay,
            )
            cv2.setMouseCallback(
                "Select coordinates: double left mouse click at two locations. Press ESC when done",
                select_cord_to_change,
            )
        if (moveStatus == True) and (insertStatus == False):
            if changeLoop == True:
                img = origImage.copy()
                changeLoop = False
            if coordChange[0] == 1:
                cv2.circle(
                    img, (cordList[2], cordList[3]), circleScale, (144, 0, 255), -1
                )
            if coordChange[0] == 2:
                cv2.circle(
                    img, (cordList[0], cordList[1]), circleScale, (144, 0, 255), -1
                )
            cv2.imshow(
                "Select coordinates: double left mouse click at two locations. Press ESC when done",
                img,
            )
            cv2.putText(
                img,
                "Click on new circle location",
                (20, 30),
                cv2.FONT_HERSHEY_TRIPLEX,
                fontScale,
                (255, 0, 255),
                2,
            )
            cv2.setMouseCallback(
                "Select coordinates: double left mouse click at two locations. Press ESC when done",
                select_new_dot_location,
            )
        if insertStatus == True:
            if coordChange[0] == 1:
                cv2.circle(
                    img, (cordList[2], cordList[3]), circleScale, (144, 0, 255), -1
                )
                cv2.circle(
                    img,
                    (newCordList[-2], newCordList[-1]),
                    circleScale,
                    (144, 0, 255),
                    -1,
                )
                cv2.line(
                    img,
                    (cordList[2], cordList[3]),
                    (newCordList[-2], newCordList[-1]),
                    (144, 0, 255),
                    int(circleScale / 4),
                )
                cordList = [newCordList[-2], newCordList[-1], cordList[2], cordList[3]]
                cordStatus = True
                moveStatus = False
                insertStatus = False
                changeLoop = True
            if coordChange[0] == 2:
                cv2.circle(
                    img, (cordList[0], cordList[1]), circleScale, (144, 0, 255), -1
                )
                cv2.circle(
                    img,
                    (newCordList[-2], newCordList[-1]),
                    circleScale,
                    (144, 0, 255),
                    -1,
                )
                cv2.line(
                    img,
                    (cordList[0], cordList[1]),
                    (newCordList[-2], newCordList[-1]),
                    (144, 0, 255),
                    int(circleScale / 4),
                )
                cordList = [cordList[0], cordList[1], newCordList[-2], newCordList[-1]]
                cordStatus = True
                moveStatus = False
                insertStatus = False
                changeLoop = True
            cv2.imshow(
                "Select coordinates: double left mouse click at two locations. Press ESC when done",
                img,
            )
        k = cv2.waitKey(20) & 0xFF
        if k == 27:
            break
    cv2.destroyAllWindows()
    os.remove(fileName)
    euclidPixelDist = np.sqrt(
        (cordList[0] - cordList[2]) ** 2 + (cordList[1] - cordList[3]) ** 2
    )
    mm_dist = int(knownmm)
    cordStatus = False
    moveStatus = False
    insertStatus = False
    changeLoop = False
    ppm = euclidPixelDist / mm_dist
    # print('pixel per mm for video ' + '"' + str(videoName) + '" = ' + str(round(ppm,4)))

    return ppm



__author__ = "Simon Nilsson"

import warnings

warnings.filterwarnings("ignore")
import glob
import math
import os
from typing import List, Optional

import numpy as np
import pandas as pd
from numba import jit, njit, prange
from scipy import stats
from scipy.signal import find_peaks, savgol_filter
from scipy.spatial import ConvexHull
from scipy.spatial.qhull import QhullError

import simba
from simba.utils.checks import (check_file_exist_and_readable,
                                check_if_filepath_list_is_empty,
                                check_minimum_roll_windows)
from simba.utils.enums import Options, Paths
from simba.utils.errors import CountError, InvalidInputError
from simba.utils.read_write import (get_bp_headers, read_config_file,
                                    read_project_path_and_file_type,
                                    read_video_info_csv)


class FeatureExtractionMixin(object):
    """
    Methods for featurizing pose-estimation data.

    :param Optional[configparser.Configparser] config_path: path to SimBA project_config.ini
    """

    def __init__(self, config_path: Optional[str] = None):

        if config_path:
            self.config_path = config_path
            self.config = read_config_file(config_path=config_path)
            self.project_path, self.file_type = read_project_path_and_file_type(
                config=self.config
            )
            self.video_info_path = os.path.join(
                self.project_path, Paths.VIDEO_INFO.value
            )
            self.video_info_df = read_video_info_csv(file_path=self.video_info_path)
            self.data_in_dir = os.path.join(
                self.project_path, Paths.OUTLIER_CORRECTED.value
            )
            self.save_dir = os.path.join(
                self.project_path, Paths.FEATURES_EXTRACTED_DIR.value
            )
            if not os.path.exists(self.save_dir):
                os.makedirs(self.save_dir)
            bp_path = os.path.join(self.project_path, Paths.BP_NAMES.value)
            check_file_exist_and_readable(file_path=bp_path)
            self.body_parts_lst = list(pd.read_csv(bp_path, header=None)[0])
            self.roll_windows_values = check_minimum_roll_windows(
                Options.ROLLING_WINDOW_DIVISORS.value, self.video_info_df["fps"].min()
            )
            self.files_found = glob.glob(self.data_in_dir + "/*." + self.file_type)
            check_if_filepath_list_is_empty(
                filepaths=self.files_found,
                error_msg=f"No files of type {self.file_type} found in {self.data_in_dir}",
            )
            self.col_headers = get_bp_headers(body_parts_lst=self.body_parts_lst)
            self.col_headers_shifted = [bp + "_shifted" for bp in self.col_headers]

    @staticmethod
    @jit(nopython=True)
    def euclidean_distance(
        bp_1_x: np.ndarray,
        bp_2_x: np.ndarray,
        bp_1_y: np.ndarray,
        bp_2_y: np.ndarray,
        px_per_mm: float,
    ) -> np.ndarray:
        """
        Helper to compute the Euclidean distance in millimeters between two body-parts in all frames of a video

        .. seealso::
           Use :meth:`simba.mixins.feature_extraction_mixin.FeatureExtractionMixin.framewise_euclidean_distance`
           for imporved run-times.

        :param np.ndarray bp_1_x: 2D array of size len(frames) x 1 with bodypart 1 x-coordinates.
        :param np.ndarray bp_2_x: 2D array of size len(frames) x 1 with bodypart 2 x-coordinates.
        :param np.ndarray bp_1_y: 2D array of size len(frames) x 1 with bodypart 1 y-coordinates.
        :param np.ndarray bp_2_y: 2D array of size len(frames) x 1 with bodypart 2 y-coordinates.
        :return np.ndarray: 2D array of size len(frames) x 1 with distances between body-part 1 and 2 in millimeters

        :example:
        >>> x1, x2 = np.random.randint(1, 10, size=(10, 1)), np.random.randint(1, 10, size=(10, 1))
        >>> y1, y2 = np.random.randint(1, 10, size=(10, 1)), np.random.randint(1, 10, size=(10, 1))
        >>> FeatureExtractionMixin.euclidean_distance(bp_1_x=x1, bp_2_x=x2, bp_1_y=y1, bp_2_y=y2, px_per_mm=4.56)
        """
        series = (np.sqrt((bp_1_x - bp_2_x) ** 2 + (bp_1_y - bp_2_y) ** 2)) / px_per_mm
        return series

    @staticmethod
    @jit(nopython=True, fastmath=True, cache=True)
    def angle3pt(
        ax: float, ay: float, bx: float, by: float, cx: float, cy: float
    ) -> float:
        """
        Jitted helper for single frame 3-point angle.

        .. seealso::
           For 3-point angles across multiple frames and improved runtime, see
           :meth:`simba.mixins.feature_extraction_mixin.FeatureExtractionMixin.angle3pt_serialized`.

        .. image:: _static/img/three_point_angle.png
           :width: 300
           :align: center

        :example:
        >>> FeatureExtractionMixin.angle3pt(ax=122.0, ay=198.0, bx=237.0, by=138.0, cx=191.0, cy=109)
        >>> 59.78156901181637
        """
        ang = math.degrees(math.atan2(cy - by, cx - bx) - math.atan2(ay - by, ax - bx))
        return ang + 360 if ang < 0 else ang

    @staticmethod
    @jit(nopython=True, fastmath=True)
    def angle3pt_serialized(data: np.ndarray) -> np.ndarray:
        """
        Jitted helper for frame-wise 3-point angles.

        .. image:: _static/img/three_point_angle.png
           :width: 300
           :align: center

        :parameter ndarray data: 2D numerical array with frame number on x and [ax, ay, bx, by, cx, cy] on y.
        :return ndarray: 1d float numerical array of size data.shape[0] with angles.

        :examples:
        >>> coordinates = np.random.randint(1, 10, size=(6, 6))
        >>> FeatureExtractionMixin.angle3pt_serialized(data=coordinates)
        >>> [ 67.16634582,   1.84761027, 334.23067238, 258.69006753, 11.30993247, 288.43494882]
        """

        results = np.full((data.shape[0]), 0.0)
        for i in prange(data.shape[0]):
            angle = math.degrees(
                math.atan2(data[i][5] - data[i][3], data[i][4] - data[i][2])
                - math.atan2(data[i][1] - data[i][3], data[i][0] - data[i][2])
            )
            if angle < 0:
                angle += 360
            results[i] = angle

        return results

    @staticmethod
    def convex_hull_calculator_mp(arr: np.ndarray, px_per_mm: float) -> float:
        """
        Calculate single frame convex hull perimeter length in millimeters.

        .. seealso::
            For acceptable run-time, call using ``parallel.delayed``.
            For large data, use :meth:`simba.feature_extractors.perimeter_jit.jitted_hull` which returns perimiter length OR area.

        .. image:: _static/img/framewise_perim_length.png
           :width: 300
           :align: center

        :parameter np.ndarray arr: 2D array of size len(body-parts) x 2.
        :parameter float px_per_mm: Video pixels per millimeter.
        :return float: The length of the animal perimeter in millimeters.

        :example:
        >>> coordinates = np.random.randint(1, 200, size=(6, 2)).astype(np.float32)
        >>> FeatureExtractionMixin.convex_hull_calculator_mp(arr=coordinates, px_per_mm=4.56)
        >>> 98.6676814218373
        """
        arr = np.unique(arr, axis=0).astype(int)
        if arr.shape[0] < 3:
            return 0
        for i in range(1, arr.shape[0]):
            if (arr[i] != arr[0]).all():
                try:
                    return ConvexHull(arr).area / px_per_mm
                except QhullError:
                    return 0
            else:
                pass
        return 0

    @staticmethod
    @jit(nopython=True)
    def count_values_in_range(data: np.ndarray, ranges: np.ndarray) -> np.ndarray:
        """
        Jitted helper finding count of values that falls within ranges. E.g., count number of pose-estimated
        body-parts that fall within defined bracket of probabilities per frame.

        .. image:: _static/img/count_ranges.png
           :width: 300
           :align: center

        :parameter np.ndarray data: 2D numpy array with frames on X.
        :parameter np.ndarray ranges: 2D numpy array representing the brackets. E.g., [[0, 0.1], [0.1, 0.5]]
        :return np.ndarray: 2D numpy array of size data.shape[0], ranges.shape[1]

        :example:
        >>> FeatureExtractionMixin.count_values_in_range(data=np.random.random((3,10)), ranges=np.array([[0.0, 0.25], [0.25, 0.5]]))
        >>> [[6, 1], [3, 2],[2, 1]]
        """

        results = np.full((data.shape[0], ranges.shape[0]), 0)
        for i in prange(data.shape[0]):
            for j in prange(ranges.shape[0]):
                lower_bound, upper_bound = ranges[j][0], ranges[j][1]
                results[i][j] = data[i][
                    np.logical_and(data[i] >= lower_bound, data[i] <= upper_bound)
                ].shape[0]
        return results

    @staticmethod
    @jit(nopython=True)
    def framewise_euclidean_distance_roi(
        location_1: np.ndarray,
        location_2: np.ndarray,
        px_per_mm: float,
        centimeter: bool = False,
    ) -> np.ndarray:
        """
        Find frame-wise distances between a moving location (location_1) and
        static location (location_2) in millimeter or centimeter.

        .. image:: _static/img/distance_to_static.png
           :width: 600
           :align: center

        :parameter ndarray location_1: 2D numpy array of size len(frames) x 2.
        :parameter ndarray location_1: 1D numpy array holding the X and Y of the static location.
        :parameter float px_per_mm: The pixels per millimeter in the video.
        :parameter bool centimeter: If true, the value in centimeters is returned. Else the value in millimeters.

        :return np.ndarray: 1D array of size location_1.shape[0]

        :example:
        >>> loc_1 = np.random.randint(1, 200, size=(6, 2)).astype(np.float32)
        >>> loc_2 = np.random.randint(1, 200, size=(1, 2)).astype(np.float32)
        >>> FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=loc_1, location_2=loc_2, px_per_mm=4.56, centimeter=False)
        >>> [11.31884926, 13.84534585,  6.09712224, 17.12773976, 19.32066031, 12.18043378]
        >>> FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=loc_1, location_2=loc_2, px_per_mm=4.56, centimeter=True)
        >>> [1.13188493, 1.38453458, 0.60971222, 1.71277398, 1.93206603, 1.21804338]
        """

        results = np.full((location_1.shape[0]), np.nan)
        for i in prange(location_1.shape[0]):
            results[i] = np.linalg.norm(location_1[i] - location_2) / px_per_mm
        if centimeter:
            results = results / 10
        return results

    @staticmethod
    @jit(nopython=True)
    def framewise_inside_rectangle_roi(
        bp_location: np.ndarray, roi_coords: np.ndarray
    ) -> np.ndarray:
        """
        Jitted helper for frame-wise analysis if animal is inside static rectangular ROI.

        .. image:: _static/img/inside_rectangle.png
           :width: 300
           :align: center

        :parameter np.ndarray bp_location:  2d numeric np.ndarray size len(frames) x 2
        :parameter np.ndarray roi_coords: 2d numeric np.ndarray size 1x2 (top left[x, y], bottom right[x, y)
        :return ndarray: 2d numeric boolean np.ndarray size len(frames) x 1 with 0 representing outside the rectangle and 1 representing inside the rectangle

        :example:
        >>> bp_loc = np.random.randint(1, 10, size=(6, 2)).astype(np.float32)
        >>> roi_coords = np.random.randint(1, 10, size=(2, 2)).astype(np.float32)
        >>> FeatureExtractionMixin.framewise_inside_rectangle_roi(bp_location=bp_loc, roi_coords=roi_coords)
        >>> [0, 0, 0, 0, 0, 0]
        """
        results = np.full((bp_location.shape[0]), 0)
        within_x_idx = np.argwhere(
            (bp_location[:, 0] <= roi_coords[1][0])
            & (bp_location[:, 0] >= roi_coords[0][0])
        ).flatten()
        within_y_idx = np.argwhere(
            (bp_location[:, 1] <= roi_coords[1][1])
            & (bp_location[:, 1] >= roi_coords[0][1])
        ).flatten()
        for i in prange(within_x_idx.shape[0]):
            match = np.argwhere(within_y_idx == within_x_idx[i])
            if match.shape[0] > 0:
                results[within_x_idx[i]] = 1
        return results

    @staticmethod
    @jit(nopython=True)
    def framewise_inside_polygon_roi(
        bp_location: np.ndarray, roi_coords: np.ndarray
    ) -> np.ndarray:
        """
        Jitted helper for frame-wise detection if animal is inside static polygon ROI.

        .. note::
           Modified from `epifanio <https://stackoverflow.com/a/48760556>`_

        .. image:: _static/img/inside_polygon.png
           :width: 300
           :align: center


        :parameter np.ndarray bp_location:  2d numeric np.ndarray size len(frames) x 2
        :parameter np.ndarray roi_coords: 2d numeric np.ndarray size len(polygon points) x 2

        :return ndarray: 2d numeric boolean np.ndarray size len(frames) x 1 with 0 representing outside the polygon and 1 representing inside the polygon

        :example:
        >>> bp_loc = np.random.randint(1, 10, size=(6, 2)).astype(np.float32)
        >>> roi_coords = np.random.randint(1, 10, size=(10, 2)).astype(np.float32)
        >>> FeatureExtractionMixin.framewise_inside_polygon_roi(bp_location=bp_loc, roi_coords=roi_coords)
        >>> [0, 0, 0, 1]
        """

        results = np.full((bp_location.shape[0]), 0)
        for i in prange(0, results.shape[0]):
            x, y, n = bp_location[i][0], bp_location[i][1], len(roi_coords)
            p2x, p2y, xints, inside = 0.0, 0.0, 0.0, False
            p1x, p1y = roi_coords[0]
            for j in prange(n + 1):
                p2x, p2y = roi_coords[j % n]
                if (
                    (y > min(p1y, p2y))
                    and (y <= max(p1y, p2y))
                    and (x <= max(p1x, p2x))
                ):
                    if p1y != p2y:
                        xints = (y - p1y) * (p2x - p1x) / (p2y - p1y) + p1x
                    if p1x == p2x or x <= xints:
                        inside = not inside
                p1x, p1y = p2x, p2y
            if inside:
                results[i] = 1

        return results

    @staticmethod
    def windowed_frequentist_distribution_tests(
        data: np.ndarray, feature_name: str, fps: int
    ) -> pd.DataFrame:
        """
        Calculates feature value distributions and feature peak counts in 1-s sequential time-bins.

        Computes (i) feature value distributions in 1-s sequential time-bins: Kolmogorov-Smirnov and T-tests.
        Computes (ii)  feature values against a normal distribution: Shapiro-Wilks.
        Computes (iii) peak count in *rolling* 1s long feature window: scipy.find_peaks.

        :parameter np.ndarray data: Single feature 1D array
        :parameter np.ndarray feature_name: The name of the input feature.
        :parameter int fps: The framerate of the video representing the data.
        :return pd.DataFrame: Of size len(data) x 4 with columns representing KS, T, Shapiro-Wilks, and peak count statistics.

        :example:
        >>> feature_data = np.random.randint(1, 10, size=(100))
        >>> FeatureExtractionMixin.windowed_frequentist_distribution_tests(data=feature_data, fps=25, feature_name='Anima_1_velocity')
        """

        (ks_results,) = (np.full((data.shape[0]), -1.0),)
        t_test_results = np.full((data.shape[0]), -1.0)
        shapiro_results = np.full((data.shape[0]), -1.0)
        peak_cnt_results = np.full((data.shape[0]), -1.0)

        for i in range(fps, data.shape[0] - fps, fps):
            bin_1_idx, bin_2_idx = [i - fps, i], [i, i + fps]
            bin_1_data, bin_2_data = (
                data[bin_1_idx[0] : bin_1_idx[1]],
                data[bin_2_idx[0] : bin_2_idx[1]],
            )
            ks_results[i : i + fps + 1] = stats.ks_2samp(
                data1=bin_1_data, data2=bin_2_data
            ).statistic
            t_test_results[i : i + fps + 1] = stats.ttest_ind(
                bin_1_data, bin_2_data
            ).statistic

        for i in range(0, data.shape[0] - fps, fps):
            shapiro_results[i : i + fps + 1] = stats.shapiro(data[i : i + fps])[0]

        rolling_idx = np.arange(fps)[None, :] + 1 * np.arange(data.shape[0])[:, None]
        for i in range(rolling_idx.shape[0]):
            bin_start_idx, bin_end_idx = rolling_idx[i][0], rolling_idx[i][-1]
            peaks, _ = find_peaks(data[bin_start_idx:bin_end_idx], height=0)
            peak_cnt_results[i] = len(peaks)

        columns = [
            f"{feature_name}_KS",
            f"{feature_name}_TTEST",
            f"{feature_name}_SHAPIRO",
            f"{feature_name}_PEAK_CNT",
        ]

        return (
            pd.DataFrame(
                np.column_stack(
                    (ks_results, t_test_results, shapiro_results, peak_cnt_results)
                ),
                columns=columns,
            )
            .round(4)
            .fillna(0)
        )

    @staticmethod
    @jit(nopython=True, cache=True)
    def cdist(array_1: np.ndarray, array_2: np.ndarray) -> np.ndarray:
        """
        Jitted analogue of meth:`scipy.cdist` for two 2D arrays. Use to calculate Euclidean distances between
        all coordinates in one array and all coordinates in a second array. E.g., computes the distances between
        all body-parts of one animal and all body-parts of a second animal.

        .. image:: _static/img/cdist.png
           :width: 600
           :align: center

        :parameter np.ndarray array_1: 2D array of body-part coordinates
        :parameter np.ndarray array_2: 2D array of body-part coordinates
        :return np.ndarray: 2D array of euclidean distances between body-parts in ``array_1`` and ``array_2``

        :example:
        >>> array_1 = np.random.randint(1, 10, size=(3, 2)).astype(np.float32)
        >>> array_2 = np.random.randint(1, 10, size=(3, 2)).astype(np.float32)
        >>> FeatureExtractionMixin.cdist(array_1=array_1, array_2=array_2)
        >>> [[7.07106781, 1.        , 3.60555124],
        >>> [3.60555124, 6.3245554 , 2.        ],
        >>>  [3.1622777 , 5.38516474, 4.12310553]])
        """
        results = np.full((array_1.shape[0], array_2.shape[0]), np.nan)
        for i in prange(array_1.shape[0]):
            for j in prange(array_2.shape[0]):
                results[i][j] = np.linalg.norm(array_1[i] - array_2[j])
        return results

    @staticmethod
    @jit(nopython=True)
    def cdist_3d(data: np.ndarray) -> np.ndarray:
        """
        Jitted analogue of meth:`scipy.cdist` for 3D array. Use to calculate Euclidean distances between
        all coordinates in of one array and itself.

        :parameter np.ndarray data: 3D array of body-part coordinates of size len(frames) x -1 x 2.
        :return np.ndarray: 3D array of size data.shape[0], data.shape[1], data.shape[1].
        """

        results = np.full((data.shape[0], data.shape[1], data.shape[1]), np.nan)
        for i in prange(data.shape[0]):
            for j in prange(data[i].shape[0]):
                for k in prange(data[i].shape[0]):
                    results[i][j][k] = np.linalg.norm(data[i][j] - data[i][k])

        return results

    @staticmethod
    # @njit('(float32[:],)')
    def cosine_similarity(data: np.ndarray) -> np.ndarray:
        """
        Jitted analogue of sklearn.metrics.pairwise import cosine_similarity. Similar to scipy.cdist.
        calculates the cosine similarity between all pairs in 2D array.

        :example:
        >>> data = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]]).astype(np.float32)
        >>> FeatureExtractionMixin().cosine_similarity(data=data)
        >>> [[1.0, 0.974, 0.959][0.974,  1.0, 0.998] [0.959, 0.998, 1.0]
        """

        dot_product = np.dot(data, data.T)
        norms = np.linalg.norm(data, axis=1).reshape(-1, 1)
        similarity = dot_product / (norms * norms.T)
        return similarity

    @staticmethod
    def create_shifted_df(df: pd.DataFrame, periods: int = 1) -> pd.DataFrame:
        """
        Create dataframe including duplicated shifted (1) columns with ``_shifted`` suffix.

        :parameter pd.DataFrame df
        :return pd.DataFrame: Dataframe including original and shifted columns.

        :example:
        >>> df = pd.DataFrame(np.random.randint(0,100,size=(3, 1)), columns=['Feature_1'])
        >>> FeatureExtractionMixin.create_shifted_df(df=df)
        >>>             Feature_1  Feature_1_shifted
        >>>    0         76               76.0
        >>>    1         41               76.0
        >>>    2         89               41.0

        """
        data_df_shifted = df.shift(periods=periods)
        data_df_shifted = data_df_shifted.combine_first(df).add_suffix("_shifted")
        return pd.concat([df, data_df_shifted], axis=1, join="inner").reset_index(
            drop=True
        )

    def check_directionality_viable(self):
        """
        Check if it is possible to calculate ``directionality`` statistics
        (i.e., nose, and ear coordinates from pose estimation has to be present)

        :return bool: If True, directionality is viable. Else, not viable.
        :return np.ndarray nose_coord: If viable, then 2D array with coordinates of the nose in all frames. Else, empty array.
        :return np.ndarray ear_left_coord: If viable, then 2D array with coordinates of the left ear in all frames. Else, empty array.
        :return np.ndarray ear_right_coord: If viable, then 2D array with coordinates of the right ear in all frames. Else, empty array.
        """

        direction_viable = True
        nose_cords, ear_left_cords, ear_right_cords = [], [], []
        for animal_name in self.animal_bp_dict.keys():
            for bp_cord in ["X_bps", "Y_bps"]:
                bp_list = self.animal_bp_dict[animal_name][bp_cord]
                for bp_name in bp_list:
                    bp_name_components = bp_name.split("_")
                    bp_name_components = [x.lower() for x in bp_name_components]
                    if "nose" in bp_name_components:
                        nose_cords.append(bp_name)
                    elif ("ear" in bp_name_components) and (
                        "left" in bp_name_components
                    ):
                        ear_left_cords.append(bp_name)
                    elif ("ear" in bp_name_components) and (
                        "right" in bp_name_components
                    ):
                        ear_right_cords.append(bp_name)
                    else:
                        pass

        for cord in [nose_cords, ear_left_cords, ear_right_cords]:
            if len(cord) != len(self.animal_bp_dict.keys()) * 2:
                direction_viable = False

        if direction_viable:
            nose_cords = [
                nose_cords[i * 2 : (i + 1) * 2]
                for i in range((len(nose_cords) + 2 - 1) // 2)
            ]
            ear_left_cords = [
                ear_left_cords[i * 2 : (i + 1) * 2]
                for i in range((len(ear_left_cords) + 2 - 1) // 2)
            ]
            ear_right_cords = [
                ear_right_cords[i * 2 : (i + 1) * 2]
                for i in range((len(ear_right_cords) + 2 - 1) // 2)
            ]

        return direction_viable, nose_cords, ear_left_cords, ear_right_cords

    def get_feature_extraction_headers(self, pose: str) -> List[str]:
        """
        Helper to return the headers names (body-part location columns) that should be used during feature extraction.

        :parameter str pose: Pose-estimation setting, e.g., ``16``.
        :return List[str]: The names and order of the pose-estimation columns.
        """
        simba_dir = os.path.dirname(simba.__file__)
        feature_categories_csv_path = os.path.join(
            simba_dir, Paths.SIMBA_FEATURE_EXTRACTION_COL_NAMES_PATH.value
        )
        check_file_exist_and_readable(file_path=feature_categories_csv_path)
        bps = list(pd.read_csv(feature_categories_csv_path)[pose])
        return [x for x in bps if str(x) != "nan"]

    @staticmethod
    @jit(nopython=True)
    def jitted_line_crosses_to_nonstatic_targets(
        left_ear_array: np.ndarray,
        right_ear_array: np.ndarray,
        nose_array: np.ndarray,
        target_array: np.ndarray,
    ) -> np.ndarray:
        """
        Jitted helper to calculate if an animal is directing towards another animals body-part coordinate,
        given the target body-part and the left ear, right ear, and nose coordinates of the observer.

        .. note::
           Input left ear, right ear, and nose coordinates of the observer is returned by
           :meth:`simba.mixins.feature_extraction_mixin.FeatureExtractionMixin.check_directionality_viable`

        .. image:: _static/img/directing_moving_targets.png
           :width: 400
           :align: center

        :parameter np.ndarray left_ear_array: 2D array of size len(frames) x 2 with the coordinates of the observer animals left ear
        :parameter np.ndarray right_ear_array: 2D array of size len(frames) x 2 with the coordinates of the observer animals right ear
        :parameter np.ndarray nose_array: 2D array of size len(frames) x 2 with the coordinates of the observer animals nose
        :parameter np.ndarray target_array: 2D array of size len(frames) x 2 with the target body-part location

        :return np.ndarray: 2D array of size len(frames) x 4. First column represent the side of the observer that the target is in view. 0 = Left side, 1 = Right side, 2 = Not in view.
        Second and third column represent the x and y location of the observer animals ``eye`` (half-way between the ear and the nose).
        Fourth column represent if target is is view (bool).

        """

        results_array = np.zeros((left_ear_array.shape[0], 4))
        for frame_no in prange(results_array.shape[0]):
            Px = np.abs(left_ear_array[frame_no][0] - target_array[frame_no][0])
            Py = np.abs(left_ear_array[frame_no][1] - target_array[frame_no][1])
            Qx = np.abs(right_ear_array[frame_no][0] - target_array[frame_no][0])
            Qy = np.abs(right_ear_array[frame_no][1] - target_array[frame_no][1])
            Nx = np.abs(nose_array[frame_no][0] - target_array[frame_no][0])
            Ny = np.abs(nose_array[frame_no][1] - target_array[frame_no][1])
            Ph = np.sqrt(Px * Px + Py * Py)
            Qh = np.sqrt(Qx * Qx + Qy * Qy)
            Nh = np.sqrt(Nx * Nx + Ny * Ny)
            if Nh < Ph and Nh < Qh and Qh < Ph:
                results_array[frame_no] = [
                    0,
                    right_ear_array[frame_no][0],
                    right_ear_array[frame_no][1],
                    True,
                ]
            elif Nh < Ph and Nh < Qh and Ph < Qh:
                results_array[frame_no] = [
                    1,
                    left_ear_array[frame_no][0],
                    left_ear_array[frame_no][1],
                    True,
                ]
            else:
                results_array[frame_no] = [2, -1, -1, False]

        return results_array

    @staticmethod
    @jit(nopython=True)
    def jitted_line_crosses_to_static_targets(
        left_ear_array: np.ndarray,
        right_ear_array: np.ndarray,
        nose_array: np.ndarray,
        target_array: np.ndarray,
    ) -> np.ndarray:
        """
        Jitted helper to calculate if an animal is directing towards a static location (ROI centroid),
        given the target location and the left ear, right ear, and nose coordinates of the observer.


        .. image:: _static/img/directing_static_targets.png
           :width: 400
           :align: center


        .. note::
           Input left ear, right ear, and nose coordinates of the observer is returned by
           :meth:`simba.mixins.feature_extraction_mixin.FeatureExtractionMixin.check_directionality_viable`

        :parameter np.ndarray left_ear_array: 2D array of size len(frames) x 2 with the coordinates of the observer animals left ear
        :parameter np.ndarray right_ear_array: 2D array of size len(frames) x 2 with the coordinates of the observer animals right ear
        :parameter np.ndarray nose_array: 2D array of size len(frames) x 2 with the coordinates of the observer animals nose
        :parameter np.ndarray target_array: 1D array of with x,y of target location

        :return np.ndarray: 2D array of size len(frames) x 4. First column represent the side of the observer that the target is in view. 0 = Left side, 1 = Right side, 2 = Not in view.
        Second and third column represent the x and y location of the observer animals ``eye`` (half-way between the ear and the nose).
        Fourth column represent if target is view (bool).

        """

        results_array = np.zeros((left_ear_array.shape[0], 4))
        for frame_no in range(results_array.shape[0]):
            Px = np.abs(left_ear_array[frame_no][0] - target_array[0])
            Py = np.abs(left_ear_array[frame_no][1] - target_array[1])
            Qx = np.abs(right_ear_array[frame_no][0] - target_array[0])
            Qy = np.abs(right_ear_array[frame_no][1] - target_array[1])
            Nx = np.abs(nose_array[frame_no][0] - target_array[0])
            Ny = np.abs(nose_array[frame_no][1] - target_array[1])
            Ph = np.sqrt(Px * Px + Py * Py)
            Qh = np.sqrt(Qx * Qx + Qy * Qy)
            Nh = np.sqrt(Nx * Nx + Ny * Ny)
            if Nh < Ph and Nh < Qh and Qh < Ph:
                results_array[frame_no] = [
                    0,
                    right_ear_array[frame_no][0],
                    right_ear_array[frame_no][1],
                    True,
                ]
            elif Nh < Ph and Nh < Qh and Ph < Qh:
                results_array[frame_no] = [
                    1,
                    left_ear_array[frame_no][0],
                    left_ear_array[frame_no][1],
                    True,
                ]
            else:
                results_array[frame_no] = [2, -1, -1, False]

        return results_array

    @staticmethod
    def minimum_bounding_rectangle(points: np.ndarray) -> np.ndarray:
        """
        Finds the minimum bounding rectangle from convex hull vertices.

        .. image:: _static/img/minimum_bounding_rectangle.png
           :width: 400
           :align: center

        :parameter np.ndarray points: 2D array representing the convexhull vertices of the animal.
        :return np.ndarray: 2D array representing minimum bounding rectangle of the convexhull vertices of the animal.

        .. note::
           Modified from `JesseBuesking <https://stackoverflow.com/questions/13542855/algorithm-to-find-the-minimum-area-rectangle-for-given-points-in-order-to-comput>`_

           See :meth:`simba.mixins.feature_extractors.perimeter_jit.jitted_hull` for computing the convexhull vertices.

        TODO: Place in numba njit.

        :example:
        >>>   points = np.random.randint(1, 10, size=(10, 2))
        >>>   FeatureExtractionMixin.minimum_bounding_rectangle(points=points)
        >>> [[10.7260274 ,  3.39726027], [ 1.4109589 , -0.09589041], [-0.31506849,  4.50684932], [ 9., 8. ]]
        """

        pi2 = np.pi / 2.0
        hull_points = points[ConvexHull(points).vertices]
        edges = hull_points[1:] - hull_points[:-1]
        angles = np.arctan2(edges[:, 1], edges[:, 0])
        angles = np.abs(np.mod(angles, pi2))
        angles = np.unique(angles)
        rotations = np.vstack(
            [np.cos(angles), np.cos(angles - pi2), np.cos(angles + pi2), np.cos(angles)]
        ).T
        rotations = rotations.reshape((-1, 2, 2))
        rot_points = np.dot(rotations, hull_points.T)
        min_x, max_x = np.nanmin(rot_points[:, 0], axis=1), np.nanmax(
            rot_points[:, 0], axis=1
        )
        min_y, max_y = np.nanmin(rot_points[:, 1], axis=1), np.nanmax(
            rot_points[:, 1], axis=1
        )
        areas = (max_x - min_x) * (max_y - min_y)
        best_idx = np.argmin(areas)
        x1, x2 = max_x[best_idx], min_x[best_idx]
        y1, y2 = max_y[best_idx], min_y[best_idx]
        r = rotations[best_idx]
        rval = np.zeros((4, 2))
        rval[0], rval[1] = np.dot([x1, y2], r), np.dot([x2, y2], r)
        rval[2], rval[3] = np.dot([x2, y1], r), np.dot([x1, y1], r)
        return rval

    @staticmethod
    @jit(nopython=True)
    def framewise_euclidean_distance(
        location_1: np.ndarray,
        location_2: np.ndarray,
        px_per_mm: float,
        centimeter: bool = False,
    ) -> np.ndarray:
        """
        Jitted helper finding frame-wise distances between two moving locations in millimeter or centimeter.

        .. image:: _static/img/framewise_euclid_dist.png
           :width: 300
           :align: center

        :parameter ndarray location_1: 2D array of size len(frames) x 2.
        :parameter ndarray location_1: 2D array of size len(frames) x 2.
        :parameter float px_per_mm: The pixels per millimeter in the video.
        :parameter bool centimeter: If true, the value in centimeters is returned. Else the value in millimeters.

        :return np.ndarray: 1D array of size location_1.shape[0]

        :example:
        >>> loc_1 = np.random.randint(1, 200, size=(6, 2)).astype(np.float32)
        >>> loc_2 = np.random.randint(1, 200, size=(6, 2)).astype(np.float32)
        >>> FeatureExtractionMixin.framewise_euclidean_distance_roi(location_1=loc_1, location_2=loc_2, px_per_mm=4.56, centimeter=False)
        >>> [49.80098657, 46.54963644, 49.60650394, 70.35919993, 37.91069901, 71.95422524]
        """
        # if not px_per_mm and centimeter:
        #     raise InvalidInputError(msg='To calculate centimeters, provide a pixel per millimeter value')
        results = np.full((location_1.shape[0]), np.nan)
        for i in prange(location_1.shape[0]):
            results[i] = np.linalg.norm(location_1[i] - location_2[i]) / px_per_mm
        if centimeter and px_per_mm:
            results = results / 10
        return results

    def change_in_bodypart_euclidean_distance(
        self,
        location_1: np.ndarray,
        location_2: np.ndarray,
        fps: int,
        px_per_mm: float,
        time_windows: np.ndarray = np.array([0.2, 0.4, 0.8, 1.6]),
    ) -> np.ndarray:
        """
        Computes the difference between the distances of two body-parts in the current frame versus N.N seconds ago.
        Used for computing if animal body-parts are traveling away or towards each other within defined time-windows.
        """
        distances = self.framewise_euclidean_distance(
            location_1=location_1, location_2=location_2, px_per_mm=px_per_mm
        )
        return self._relative_distances(
            distances=distances, fps=fps, time_windows=time_windows
        )

    def dataframe_gaussian_smoother(
        self, df: pd.DataFrame, fps: int, time_window: int = 100
    ) -> pd.DataFrame:
        """
        Column-wise Gaussian smoothing of dataframe.

        :parameter pd.DataFrame df: Dataframe with un-smoothened data.
        :parameter int fps: The frame-rate of the video representing the data.
        :parameter int time_window: Time-window in milliseconds to use for Gaussian smoothing.
        :return pd.DataFrame: Dataframe with smoothened data

        :references:
        .. [1] `Video expected putput <https://www.youtube.com/watch?v=d9-Bi4_HyfQ>`__.
        """

        frames_in_time_window = int(time_window / (1000 / fps))
        for c in df.columns:
            df[c] = (
                df[c]
                .rolling(
                    window=int(frames_in_time_window), win_type="gaussian", center=True
                )
                .mean(std=5)
                .fillna(df[c])
                .abs()
                .astype(int)
            )
        return df

    def dataframe_savgol_smoother(
        self, df: pd.DataFrame, fps: int, time_window: int = 150
    ) -> pd.DataFrame:
        """
        Column-wise Savitzky-Golay smoothing of dataframe.

        :parameter pd.DataFrame df: Dataframe with un-smoothened data.
        :parameter int fps: The frame-rate of the video representing the data.
        :parameter int time_window: Time-window in milliseconds to use for Gaussian smoothing.
        :return pd.DataFrame: Dataframe with smoothened data

        :references:
        .. [1] `Video expected putput <https://www.youtube.com/watch?v=d9-Bi4_HyfQ>`__.

        """
        frames_in_time_window = int(time_window / (1000 / fps))
        if (frames_in_time_window % 2) == 0:
            frames_in_time_window = frames_in_time_window - 1
        if (frames_in_time_window % 2) <= 3:
            frames_in_time_window = 5
        for c in df.columns:
            df[c] = savgol_filter(
                x=df[c].to_numpy(),
                window_length=frames_in_time_window,
                polyorder=3,
                mode="nearest",
            ).astype(int)
        return df

    def get_bp_headers(self) -> None:
        """
        Helper to create ordered list of all column header fields for SimBA project dataframes.
        """
        self.col_headers = []
        for bp in self.body_parts_lst:
            c1, c2, c3 = (f"{bp}_x", f"{bp}_y", f"{bp}_p")
            self.col_headers.extend((c1, c2, c3))

    def check_directionality_cords(self) -> dict:
        """
        Helper to check if ear and nose body-parts are present within the pose-estimation data.

        :return dict: Body-part names of ear and nose body-parts as values and animal names as keys. If empty,
            ear and nose body-parts are not present within the pose-estimation data

        """
        results = {}
        for animal in self.animal_bp_dict.keys():
            results[animal] = {}
            results[animal]["Nose"] = {}
            results[animal]["Ear_left"] = {}
            results[animal]["Ear_right"] = {}
            for dimension in ["X_bps", "Y_bps"]:
                for cord in self.animal_bp_dict[animal][dimension]:
                    if ("nose" in cord.lower()) and ("x" in cord.lower()):
                        results[animal]["Nose"]["X_bps"] = cord
                    elif ("nose" in cord.lower()) and ("y" in cord.lower()):
                        results[animal]["Nose"]["Y_bps"] = cord
                    elif (
                        ("left" in cord.lower())
                        and ("x" in cord.lower())
                        and ("ear" in cord.lower())
                    ):
                        results[animal]["Ear_left"]["X_bps"] = cord
                    elif (
                        ("left" in cord.lower())
                        and ("Y".lower() in cord.lower())
                        and ("ear".lower() in cord.lower())
                    ):
                        results[animal]["Ear_left"]["Y_bps"] = cord
                    elif (
                        ("right" in cord.lower())
                        and ("x" in cord.lower())
                        and ("ear" in cord.lower())
                    ):
                        results[animal]["Ear_right"]["X_bps"] = cord
                    elif (
                        ("right" in cord.lower())
                        and ("y" in cord.lower())
                        and ("ear".lower() in cord.lower())
                    ):
                        results[animal]["Ear_right"]["Y_bps"] = cord
        return results

    def insert_default_headers_for_feature_extraction(
        self, df: pd.DataFrame, headers: List[str], pose_config: str, filename: str
    ) -> pd.DataFrame:
        """
        Helper to insert correct body-part column names prior to defualt feature extraction methods.
        """
        if len(headers) != len(df.columns):
            raise CountError(
                f"Your SimBA project is set to using the default {pose_config} pose-configuration. "
                f"SimBA therefore expects {str(len(headers))} columns of data inside the files within the project_folder. However, "
                f"within file {filename} file, SimBA found {str(len(df.columns))} columns.",
                source=self.__class__.__name__,
            )
        else:
            df.columns = headers
            return df

    @staticmethod
    def line_crosses_to_static_targets(
        p: List[float],
        q: List[float],
        n: List[float],
        M: List[float],
        coord: List[float],
    ) -> (bool, List[float]):
        """
        Legacy non-jitted helper to calculate if an animal is directing towards a static coordinate (e.g., ROI centroid).

        .. note:
           For improved runtime, use :meth:`simba.mixins.feature_extraction_mixin.jitted_line_crosses_to_static_targets`

        :parameter list p: left ear coordinates of observing animal.
        :parameter list q: right ear coordinates of observing animal.
        :parameter list n: nose coordinates of observing animal.
        :parameter list M: The location of the target coordinates.
        :parameter list coord: empty list to store the eye coordinate of the observing animal.

        :return bool: If True, static coordinate is in view.
        :return List: If True, the coordinate of the observing animals ``eye`` (half-way between nose and ear).
        """

        Px = np.abs(p[0] - M[0])
        Py = np.abs(p[1] - M[1])
        Qx = np.abs(q[0] - M[0])
        Qy = np.abs(q[1] - M[1])
        Nx = np.abs(n[0] - M[0])
        Ny = np.abs(n[1] - M[1])
        Ph = np.sqrt(Px * Px + Py * Py)
        Qh = np.sqrt(Qx * Qx + Qy * Qy)
        Nh = np.sqrt(Nx * Nx + Ny * Ny)
        if Nh < Ph and Nh < Qh and Qh < Ph:
            coord.extend((q[0], q[1]))
            return True, coord
        elif Nh < Ph and Nh < Qh and Ph < Qh:
            coord.extend((p[0], p[1]))
            return True, coord
        else:
            return False, coord

    @staticmethod
    @njit("(int64[:,:], int64[:,:], float64)")
    def find_midpoints(
        bp_1: np.ndarray, bp_2: np.ndarray, percentile: float = 0.5
    ) -> np.ndarray:
        """
        Compute the midpoints between two sets of 2D points based on a given percentile.

        :parameter np.ndarray bp_1: An array of 2D points representing the first set of points. Rows represent frames. First column represent x coordinates. Second column represent y coordinates.
        :parameter np.ndarray bp_2: An array of 2D points representing the second set of points. Rows represent frames. First column represent x coordinates. Second column represent y coordinates.
        :parameter float percentile: The percentile value to determine the distance between the points for calculating midpoints. When set to 0.5 it calculates midpoints at the midpoint of the two points.
        :returns: np.ndarray: An array of 2D points representing the midpoints between the points in bp_1 and bp_2 based on the specified percentile.


        .. image:: _static/img/find_midpoints.png
           :width: 600
           :align: center

        :example:
        >>> bp_1 = np.array([[1, 3], [30, 10]]).astype(np.int64)
        >>> bp_2 = np.array([[10, 4], [20, 1]]).astype(np.int64)
        >>> FeatureExtractionMixin().find_midpoints(bp_1=bp_1, bp_2=bp_2, percentile=0.5)
        >>> [[ 5,  3], [25,  6]]
        """

        result = np.full(bp_1.shape, np.nan)
        for i in range(bp_1.shape[0]):
            frm_bp1, frm_bp2 = bp_1[i], bp_2[i]
            x_dist = max(frm_bp1[0], frm_bp2[0]) - min(frm_bp1[0], frm_bp2[0])
            y_dist = max(frm_bp1[1], frm_bp2[1]) - min(frm_bp1[1], frm_bp2[1])
            x_dist_percentile, y_dist_percentile = int(x_dist * percentile), int(
                y_dist * percentile
            )
            if frm_bp2[0] > frm_bp1[0]:
                new_x = frm_bp1[0] + x_dist_percentile
            else:
                new_x = frm_bp1[0] - x_dist_percentile
            if frm_bp2[1] > frm_bp1[1]:
                new_y = frm_bp1[1] + y_dist_percentile
            else:
                new_y = frm_bp1[1] - y_dist_percentile

            result[i] = np.array([new_x, new_y])

        return result.astype(np.int64)
